from joule.models.pipes.pipe import Pipe, interval_token, find_interval_token, compute_dtype
from joule.models.pipes.input_pipe import InputPipe
from joule.models.pipes.output_pipe import OutputPipe
from joule.models.pipes.local_pipe import LocalPipe
from joule.models.pipes.factories import reader_factory, writer_factory
from joule.models.pipes.errors import EmptyPipe, PipeError



from joule.client.reader_module import ReaderModule
from joule.client.filter_module import FilterModule
from joule.client.fir_filter_module import FIRFilterModule
from joule.client.composite_module import CompositeModule
from joule.client.base_module import BaseModule


import click
from .copy import data_copy
from .read import cmd as data_read
from .remove import data_remove
from .intervals import intervals
from .consolidate import consolidate
from .merge import merge
from .mean import mean
from .median import median
from .ingest import ingest

@click.group(name="data")
def data():
    """Interact with data streams."""
    pass  # pragma: no cover

@click.group(name="filter")
def filter():
    """Filter stream data."""
    pass  # pragma: no cover

data.add_command(data_copy)
data.add_command(merge)
data.add_command(data_read)
data.add_command(data_remove)
data.add_command(intervals)
data.add_command(consolidate)
data.add_command(ingest)
data.add_command(filter)

filter.add_command(mean)
filter.add_command(median)



import argparse
import asyncio
import signal

from typing import List, Tuple, Dict, Optional
from aiohttp import web
import socket
import logging
import os
import json

import joule.utilities
from joule.api import node
from joule import api
from joule.api.data_stream import DataStream, Element
from joule.client import helpers
# import directly so it can be mocked easily in unit tests
from joule.errors import ConfigurationError, EmptyPipeError, ApiError
from joule.models import pipes
from joule.services.parse_pipe_config import parse_pipe_config, parse_inline_config
from joule.utilities.interval_tools import interval_intersection, interval_difference

Pipes = Dict[str, 'pipes.Pipe']
Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')

DataStreams = Dict[str, api.DataStream]


class BaseModule:
    """
    This is an abstract class and should not be inherited directly. Instead inherit from one of
    the following children :class:`joule.ReaderModule`, :class:`joule.FilterModule`, or :class:`joule.CompositeModule`
    """

    def __init__(self):
        self.stop_requested = False
        self.pipes: List[pipes.Pipe] = []
        self.STOP_TIMEOUT = 2
        self.runner = None
        self.node: node.BaseNode = None

    async def run_as_task(self, parsed_args, app: web.Application) -> asyncio.Task:
        assert False, "implement in child class"  # pragma: no cover

    def custom_args(self, parser: argparse.ArgumentParser):
        """

        Override to add custom command line arguments to the module.

        .. code-block:: python

            class ModuleDemo(BaseModule):

                def custom_args(self, parser):
                     parser.description = "**module description**"
                     # add optional help text to the argument
                     parser.add_argument("--arg", help="custom argument")
                     # parse json input
                     parser.add_argument("--json_arg", type=json.loads)
                     # a yes|no argument that resolves to True|False
                     parser.add_argument("--flag_arg", type=joule.yesno)

                #... other module code


        Always use keyword arguments with modules so they can be specified
        in the **[Arguments]** section  of module configuration file

        Use the ``type`` parameter to specify a parser function. The parser
        function should accept a string input and return the appropriate object.

        """
        # parser.add_argument("--custom_flag")
        pass  # pragma: no cover

    def stop(self):
        """
        Override to change the default shutdown strategy which simply sets
        the ``stop_requested`` flag. If a module does not terminate within a few seconds
        of this method being called Joule will forcibly stop the module with SIGKILL.
        """
        # override in client for alternate shutdown strategy
        self.stop_requested = True

    def create_dev_app(self) -> web.Application:
        parser = argparse.ArgumentParser()
        self._build_args(parser)
        # must specify module config in JOULE_MODULE_CONFIG environment variable
        if 'JOULE_MODULE_CONFIG' not in os.environ:
            raise ConfigurationError("JOULE_MODULE_CONFIG not set, must specify config file")
        module_config_file = os.environ['JOULE_MODULE_CONFIG']
        if not os.path.isfile(module_config_file):
            raise ConfigurationError("JOULE_MODULE_CONFIG is not a module config file")

        module_args = helpers.load_args_from_file(module_config_file)
        parsed_args = parser.parse_args(module_args)
        self.node = api.get_node(parsed_args.node)

        self.stop_requested = False
        my_app = self._create_app()

        async def on_startup(app):
            app['task'] = await self.run_as_task(parsed_args, app)

        my_app.on_startup.append(on_startup)

        async def on_shutdown(app):
            self.stop()
            await app['task']

        my_app.on_shutdown.append(on_shutdown)
        return my_app

    def start(self, parsed_args: argparse.Namespace = None):
        """
        Execute the module. Do not override this function. Creates an event loop and
        executes the :meth:`run` coroutine.

        Args:
            parsed_args: omit to parse the command line arguments

        .. code-block:: python

            class ModuleDemo(BaseModule):
                # body of module...
                # at a minimum the run coroutine must be implemented

            if __name__ == "__main__":
                my_module = ModuleDemo()
                my_module.start()
        """

        if parsed_args is None:  # pragma: no cover
            parser = argparse.ArgumentParser()
            self._build_args(parser)
            module_args = helpers.module_args()
            parsed_args = parser.parse_args(module_args)

        # asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        loop = asyncio.get_event_loop()
        self.stop_requested = False
        if parsed_args.api_socket != "unset":
            # should be set by the joule daemon
            if 'JOULE_CA_FILE' in os.environ:
                cafile = os.environ['JOULE_CA_FILE']
            else:
                cafile = ""
            self.node = node.UnixNode("local", parsed_args.api_socket, cafile)
        else:
            self.node = api.get_node(parsed_args.node)
        self.runner: web.AppRunner = loop.run_until_complete(self._start_interface(parsed_args))
        app = None
        if self.runner is not None:
            app = self.runner.app
        try:
            task = loop.run_until_complete(
                self.run_as_task(parsed_args, app))
        except ConfigurationError as e:
            log.error("ERROR: " + str(e))
            self._cleanup(loop)
            return

        def stop_task():
            # give task no more than 2 seconds to exit
            loop.call_later(self.STOP_TIMEOUT, task.cancel)
            # run custom exit routine
            self.stop()

        loop.add_signal_handler(signal.SIGINT, stop_task)
        loop.add_signal_handler(signal.SIGTERM, stop_task)
        try:
            loop.run_until_complete(task)
        except (asyncio.CancelledError, pipes.EmptyPipe, EmptyPipeError) as e:
            pass
        finally:
            self._cleanup(loop)

    def _cleanup(self, loop: Loop):
        if self.runner is not None:
            loop.run_until_complete(self.runner.cleanup())
        for pipe in self.pipes:
            loop.run_until_complete(pipe.close())
        if self.node is not None:
            loop.run_until_complete(self.node.close())
        loop.run_until_complete(asyncio.sleep(0))
        loop.close()

    def _build_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group('joule',
                                        'control module execution')
        # --pipes: JSON argument set by jouled
        grp.add_argument("--pipes",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --socket: UNIX socket set by jouled for interface proxy
        grp.add_argument("--socket",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --api-socket: UNIX socket for API calls back to jouled
        grp.add_argument("--api-socket",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --port: port to host interface on during isolated execution
        grp.add_argument("--port", type=int,
                         default=8080,
                         help='port for isolated execution')
        # --host: IP address to host interface on during isolated exedcution
        grp.add_argument("--host",
                         default="0.0.0.0",
                         help="IP address for isolated execution")
        # --module_config: set to run module standalone
        grp.add_argument("--module_config",
                         default="unset",
                         help='specify *.conf file for isolated execution')
        # --stream_configs: set to run module standalone
        grp.add_argument("--stream_configs",
                         default="unset",
                         help="specify directory of stream configs " +
                              "for isolated execution")
        # --node: node to connect to for isolated execution
        grp.add_argument("--node", default="",
                         help="joule node for isolated execution")
        # --new: historical isolation mode
        grp.add_argument("--new", action="store_true",
                         help="process historic data newer than the most recent destination data")
        # --live: subscribe to inputs (inputs must be actively produced)
        grp.add_argument("--live", action="store_true",
                         help="process live input data")
        # --start_time: historical isolation mode
        grp.add_argument("--start_time", default=None,
                         help="input start time for historic isolation")
        # --end_time: historical isolation mode
        grp.add_argument("--end_time", default=None,
                         help="input end time for historic isolation")
        # --force: do not ask for confirmation
        grp.add_argument("--force", action="store_true",
                         help="do not ask for confirmation before dangerous operations")

        parser.formatter_class = argparse.RawDescriptionHelpFormatter
        self.custom_args(parser)

    async def _parse_streams(self, parsed_args) -> Tuple[DataStreams, DataStreams]:
        inputs = {}
        outputs = {}
        # if this is running in Joule use the pipe args structure with API calls
        if parsed_args.pipes != 'unset':
            try:
                pipe_json = json.loads(json.loads(parsed_args.pipes))
                # if debugging, pycharm escapes the outer JSON
                # pipe_json = json.loads(pipe_args.encode('utf-8').decode('unicode_escape'))
                output_args = pipe_json['outputs']
                input_args = pipe_json['inputs']
                for name, config in input_args.items():
                    if config['id'] is None:  # remote stream, no information available
                        inputs[name] = self._create_stub_stream(name, config['layout'])
                    else:
                        inputs[name] = await self.node.data_stream_get(config['id'])
                for name, config in output_args.items():
                    if config['id'] is None:  # remote stream, no information available
                        outputs[name] = self._create_stub_stream(name, config['layout'])
                    else:
                        outputs[name] = await self.node.data_stream_get(config['id'])
            except (KeyError, json.JSONDecodeError):
                raise ConfigurationError(f"invalid pipes argument: {parsed_args.pipes}")
        # otherwise use the specified module configuration file
        else:
            if parsed_args.module_config == "unset":
                raise ConfigurationError("module_config parameter missing")
            module_config = helpers.read_module_config(parsed_args.module_config)
            if parsed_args.stream_configs != "unset":
                configured_streams = helpers.read_stream_configs(parsed_args.stream_configs)
            else:
                configured_streams = {}
            if 'Inputs' in module_config:
                for name, path in module_config['Inputs'].items():
                    stream = await self._validate_or_create_stream(configured_streams, path)
                    inputs[name] = stream
            if 'Outputs' in module_config:
                for name, path in module_config['Outputs'].items():
                    stream = await self._validate_or_create_stream(configured_streams, path)
                    outputs[name] = stream
        return inputs, outputs

    def _create_stub_stream(self, name, layout):
        # if API information is not available (coming from a remote node) then
        # create a stream with the right layout to use as a substitute
        # NOTE: this could be improved but since the remote capability isn't
        # really used this should be fine for now
        (dtype, num_elements) = layout.split('_')
        return DataStream(name=f"STUB:{name}", description="automatically generated stub",
                          datatype=dtype,
                          elements=[Element(name=f"E{i}") for i in range(int(num_elements))])

    async def _validate_or_create_stream(self, configured_streams, raw_path):
        # process inline configuration if present
        path = await self._process_inline_configuration(raw_path, configured_streams)
        # try to get the actual stream from Joule
        try:
            stream = await self.node.data_stream_get(path)
            # validate that the local configuration matches the actual one
            if path in configured_streams:
                if configured_streams[path].layout != stream.layout:
                    raise ConfigurationError("Stream [%s] exists with layout [%s], not [%s]" % (
                        path, stream.layout, configured_streams[path].layout
                    ))
            return stream
        except ApiError:
            pass
        # create the stream
        if path not in configured_streams:
            raise ConfigurationError(f"{path} does not exist, provide config file or inline configuration")
        folder_path = '/'.join(path.split('/')[:-1])
        return await self.node.data_stream_create(configured_streams[path], folder_path)

    async def _process_inline_configuration(self, pipe_config, configured_streams):
        # if there is an inline configuration add it to the configured streams dict
        # returns the full path to the stream
        (path, name, inline_config) = parse_pipe_config(pipe_config)
        if inline_config == '':
            return pipe_config  # the pipe config is just the full path
        full_path = "/".join([path, name])
        # inline config exists
        (datatype, element_names) = parse_inline_config(inline_config)
        datatype = datatype.name.lower()  # API models are plain text attributes
        # make sure inline config agrees with stream config if present
        if full_path in configured_streams:
            stream = configured_streams[full_path]
            if datatype != stream.datatype.lower() or \
                    len(stream.elements) != len(element_names):
                raise ConfigurationError(
                    f"Invalid configuration: [{name}] inline format does not match config file")
        # otherwise create a stream object from the inline config
        stream = DataStream()
        stream.name = name
        stream.decimate = True
        stream.datatype = datatype
        for i in range(len(element_names)):
            e = Element()
            e.name = element_names[i]
            e.index = i
            stream.elements.append(e)
        configured_streams[full_path] = stream
        return full_path

    async def _compute_missing_intervals(self, input_streams, output_streams, parsed_args):
        # 1) If live is specified, don't do anything else- no intervals to process
        if parsed_args.live or parsed_args.pipes != 'unset':
            return [None]
        # 2) Convert string time arguments into Optional[timestamp] types
        start, end = helpers.validate_time_bounds(parsed_args.start_time,
                                                  parsed_args.end_time)
        # 3) Find the effective input and output intervals (intersection of the streams)
        input_intervals = None
        for stream in input_streams.values():
            intervals = await self.node.data_intervals(stream, start, end)
            if input_intervals is not None:
                input_intervals = interval_intersection(input_intervals, intervals)
            else:
                input_intervals = intervals
        output_intervals = None
        for stream in output_streams.values():
            intervals = await self.node.data_intervals(stream, start, end)
            if output_intervals is not None:
                output_intervals = interval_intersection(output_intervals, intervals)
            else:
                output_intervals = intervals
        if output_intervals is None:
            missing_intervals = input_intervals
        else:
            missing_intervals = interval_difference(input_intervals, output_intervals)
        # filter out intervals shorter than 1 second (boundary error with intervals)
        missing_intervals = [i for i in missing_intervals if (i[1] - i[0]) > 1e6]
        return missing_intervals

    async def _build_pipes_new(self, interval, input_streams, output_streams, pipe_args) -> Tuple[Pipes, Pipes]:
        input_pipes = {}
        output_pipes = {}
        # use network sockets for connection to inputs and outputs
        if pipe_args == 'unset':
            for (name, stream) in input_streams.items():
                if interval is None:  # subscribe to live data
                    input_pipes[name] = await self.node.data_subscribe(stream)
                else:
                    input_pipes[name] = await self.node.data_read(stream, interval[0], interval[1])
            for (name, stream) in output_streams.items():
                if interval is None:
                    output_pipes[name] = await self.node.data_write(stream)
                else:
                    output_pipes[name] = await self.node.data_write(stream, interval[0], interval[1])
        # use file descriptors provided by joule for connection to inputs and outputs
        else:
            try:
                pipe_json = json.loads(json.loads(pipe_args))
                # if debugging, pycharm escapes the outer JSON
                # pipe_json = json.loads(pipe_args.encode('utf-8').decode('unicode_escape'))
                output_args = pipe_json['outputs']
                input_args = pipe_json['inputs']
            except (KeyError, json.JSONDecodeError):
                raise ConfigurationError(f"invalid pipes argument: {pipe_args}")

            for name, arg in output_args.items():
                wf = pipes.writer_factory(arg['fd'])
                output_pipes[name] = pipes.OutputPipe(stream=output_streams[name],
                                                      layout=arg['layout'],
                                                      writer_factory=wf)
            for name, arg in input_args.items():
                rf = pipes.reader_factory(arg['fd'])
                input_pipes[name] = pipes.InputPipe(stream=input_streams[name],
                                                    layout=arg['layout'],
                                                    reader_factory=rf)
        # keep track of the pipes so they can be closed
        self.pipes = list(input_pipes.values()) + list(output_pipes.values())
        return input_pipes, output_pipes

    async def _build_pipes(self, parsed_args) -> Tuple[Pipes, Pipes]:
        # for unit testing mocks
        from joule.client.helpers.pipes import build_network_pipes

        pipe_args = parsed_args.pipes

        # figure out whether we should run with fd's or network sockets
        if pipe_args == 'unset':
            if parsed_args.module_config == 'unset':
                raise ConfigurationError('must specify --module_config')

            # 1) Convert string time arguments into Optional[timestamp] types
            start, end = helpers.validate_time_bounds(parsed_args.start_time,
                                                      parsed_args.end_time)
            # 2) Parse module config to find input and output streams
            inputs = {}
            outputs = {}
            module_config = helpers.read_module_config(parsed_args.module_config)
            if 'Inputs' in module_config:
                for name, path in module_config['Inputs'].items():
                    inputs[name] = path
            if 'Outputs' in module_config:
                for name, path in module_config['Outputs'].items():
                    outputs[name] = path
            # 3) Parse all stream config files in case they are needed
            configured_streams = {}
            if parsed_args.stream_configs != "unset":
                configured_streams = helpers.read_stream_configs(parsed_args.stream_configs)
            # 4) Build the input and output pipes using network sockets
            pipes_in, pipes_out = await build_network_pipes(
                inputs, outputs, configured_streams, self.node, start, end,
                parsed_args.new, parsed_args.force)
        else:
            # run the module within joule
            (pipes_in, pipes_out) = await helpers.build_fd_pipes(pipe_args, self.node)
        # keep track of the pipes so they can be closed
        self.pipes = list(pipes_in.values()) + list(pipes_out.values())
        return pipes_in, pipes_out

    def routes(self):
        """
        Override to register HTTP handlers for the module. Return
        an array of handlers. This creates a visualization interface.

        .. code-block:: python

            class ModuleDemo(BaseModule):

                def routes(self):
                    return [
                        web.get('/', self.index),
                        # other handlers ...
                    ]

                async def index(self, request):
                    return web.Response(text="Hello World")

                #... other module code

        """
        return []  # override in child to implement a web interface

    def _create_app(self) -> web.Application:
        routes = self.routes()
        app = web.Application()
        app.add_routes(routes)
        return app

    async def _start_interface(self, args) -> Optional[web.AppRunner]:
        routes = self.routes()
        # socket config is 'unset' when run as a standalone process
        if args.socket == 'unset':
            if len(routes) == 0:
                return None  # no routes requested, nothing to do
            # create a local server
            app = web.Application()
            app.add_routes(routes)
            runner = web.AppRunner(app)
            await runner.setup()
            site = web.TCPSite(runner, port=args.port, host=args.host)
            await site.start()
            print("starting web server at %s:%d" % (args.host, args.port))
            return runner
        # socket config is 'none' when joule does not connect a socket
        if args.socket == 'none':
            if len(routes) > 0:
                logging.error("No socket available for the interface, check module configuration")
            return None
        # otherwise start a UNIX runner on the socket
        if os.path.exists(args.socket):
            log.error("Socket address [%s] is already in use, cannot start interface" % socket)
            return None
        app = web.Application()
        app.add_routes(routes)
        runner = web.AppRunner(app)
        await runner.setup()
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(args.socket)
        site = web.SockSite(runner, sock)
        await site.start()
        print("starting web server at [%s]" % args.socket)
        return runner



from typing import List, Callable, Dict, Optional
import asyncio
import yarl
import logging
from sqlalchemy import orm

from joule.models import Worker, DataStream, Proxy, Follower, pipes
from joule.errors import SubscriptionError, ConfigurationError, ApiError

from joule.api import BaseNode

Tasks = List[asyncio.Task]
Loop = asyncio.AbstractEventLoop

log = logging.getLogger('joule')


class Supervisor:

    def __init__(self, workers: List[Worker], proxies: List[Proxy], get_node: Callable[[str], BaseNode]):
        self._workers = workers
        self._proxies = proxies
        self.get_node = get_node
        self.task: asyncio.Task = None
        self.remote_tasks: List[asyncio.Task] = []
        self.remote_inputs: Dict[DataStream, pipes.Pipe] = {}

        self.REMOTE_HANDLER_RESTART_INTERVAL = 5

    @property
    def workers(self) -> List[Worker]:
        return self._workers

    @property
    def proxies(self) -> List[Proxy]:
        return self._proxies

    async def start(self):
        # returns a co-routine
        tasks: Tasks = []
        for worker in self._workers:
            await self._connect_remote_outputs(worker)
            t = asyncio.create_task(worker.run(self.subscribe))
            t.set_name("Supervising worker [%s]" % worker.name)
            tasks.append(t)
        self.task = asyncio.gather(*tasks)

    async def stop(self):
        for worker in self._workers:
            await worker.stop()
        try:
            await self.task
        except Exception as e:
            log.warning("Supervisor worker shutdown exception: %s " % str(e))
        for task in self.remote_tasks:
            task.cancel()
            try:
                await task
            except Exception as e:
                log.warning("Supervisor remote i/o shutdown exception: %s " % str(e))
                raise e

    async def restart_producer(self, stream: DataStream, msg=""):
        # find the worker who produces this stream
        for worker in self._workers:
            if worker.produces(stream):
                if msg is not None:
                    log.warning("Restarting module [%s]: %s" % (worker.name, msg))
                    worker.log("[Supervisor Restarting Module: %s]" % msg)
                await worker.restart()

    def subscribe(self, stream: DataStream, pipe: pipes.Pipe) -> Callable:
        # if the stream is remote, connect to it
        if stream.is_remote:
            return self._connect_remote_input(stream, pipe)
        # otherwise find a worker producing it
        for worker in self._workers:
            try:
                return worker.subscribe(stream, pipe)
            except SubscriptionError:
                pass
        else:
            raise SubscriptionError("stream [%s] has no producer" % stream.name)

    def get_module_socket(self, uuid):
        for w in self.workers:
            if w.uuid == uuid:
                return w.interface_socket
        return None

    def get_proxy_url(self, uuid) -> Optional[yarl.URL]:
        for p in self._proxies:
            if p.uuid == uuid:
                return p.url
        return None

    async def _connect_remote_outputs(self, worker: Worker):
        """ Provide a pipe to the worker for each remote stream, spawn a
            task that reads from the worker's pipe and writes out to a remote
            network pipe, if the remote network pipe goes down or is not available,
            continuously try to restore it
        """
        remote_streams = [stream for stream in worker.subscribers if stream.is_remote]
        for stream in remote_streams:
            src_pipe = pipes.LocalPipe(stream.layout, stream=stream)
            # ignore unsubscribe cb, not used
            worker.subscribe(stream, src_pipe)
            task = asyncio.create_task(self._handle_remote_output(src_pipe, stream))
            self.remote_tasks.append(task)

    async def _handle_remote_output(self, src_pipe: pipes.Pipe, dest_stream: DataStream):
        """
        Continuously tries to make a connection to dest_stream and write src_pipe's data to it
        """
        dest_pipe = None
        node = self.get_node(dest_stream.remote_node)
        if node is None:
            log.error("output requested from [%s] but this node is not a follower" % dest_stream.remote_node)
            return
        try:
            while True:
                try:
                    dest_pipe = await node.data_write(dest_stream.remote_path)
                    while True:
                        data = await src_pipe.read()
                        await dest_pipe.write(data)
                        src_pipe.consume(len(data))

                except ConfigurationError as e:
                    log.error("Subscriber::_handle_remote_output: %s" % str(e))
                    await asyncio.sleep(self.REMOTE_HANDLER_RESTART_INTERVAL)
        except asyncio.CancelledError:
            pass
        finally:
            await src_pipe.close()
            if dest_pipe is not None:
                await dest_pipe.close()
            await node.close()

    def _connect_remote_input(self, stream: DataStream, pipe: pipes.Pipe):
        """
        Spawn a task that maintains a connection with [stream] and provides the data to [pipe]
        """
        # somebody is already listening to stream, just subscribe to that pipe
        if stream in self.remote_inputs:
            return self.remote_inputs[stream].subscribe(pipe)

        # this is the first subscriber, spawn an input task to feed the pipe
        task = asyncio.create_task(self._handle_remote_input(stream, pipe))

        self.remote_inputs[stream] = pipe
        self.remote_tasks.append(task)

    async def _handle_remote_input(self, src_stream: DataStream, dest_pipe: pipes.Pipe):
        """
        Continuously tries to make a connection to src_stream and write its data to dest_pipe
        """
        src_pipe = None
        node = self.get_node(src_stream.remote_node)
        if node is None:
            log.error("input requested from [%s] but this node is not a follower" % src_stream.remote_node)
            return

        try:
            while True:
                try:
                    src_pipe = await node.data_subscribe(src_stream.remote_path)
                    try:
                        while True:
                            data = await src_pipe.read()
                            if not dest_pipe.closed:
                                await dest_pipe.write(data)
                            else:
                                log.info(
                                    "destination [%s] is closed, OK if this is during shutdown" % dest_pipe.stream.name)
                            src_pipe.consume(len(data))
                    except pipes.EmptyPipe:
                        await dest_pipe.close_interval()
                    log.error("Subscriber:: _handle_remote_input: connection terminated unexepectedly")
                except (ConfigurationError, ApiError) as e:
                    log.error("Subscriber::_handle_remote_input: %s" % str(e))
                await asyncio.sleep(self.REMOTE_HANDLER_RESTART_INTERVAL)
        except asyncio.CancelledError:
            pass
        finally:
            if src_pipe is not None:
                await src_pipe.close()
            await dest_pipe.close()
            await node.close()



from sqlalchemy.orm import Session
from aiohttp import web
import numpy as np
import asyncio
import logging
import aiohttp


from joule.models import (folder, DataStore, DataStream,
                          InsufficientDecimationError, DataError,
                          pipes)
from joule.models.supervisor import Supervisor

from joule.errors import SubscriptionError

log = logging.getLogger('joule')


async def read_json(request: web.Request):
    return await read(request, json=True)


async def read(request: web.Request, json=False):
    if 'subscribe' in request.query and request.query['subscribe'] == '1':
        return await _subscribe(request, json)
    else:
        return await _read(request, json)


async def _read(request: web.Request, json):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse optional parameters
    params = {'start': None, 'end': None, 'max-rows': None, 'decimation-level': None}
    param = ""  # to appease type checker

    try:
        for param in params:
            if param in request.query:
                params[param] = int(request.query[param])
    except ValueError:
        return web.Response(text="parameter [%s] must be an int" % param, status=400)

    # make sure parameters make sense
    if ((params['start'] is not None and params['end'] is not None) and
            (params['start'] >= params['end'])):
        return web.Response(text="[start] must be < [end]", status=400)
    if params['max-rows'] is not None and params['max-rows'] <= 0:
        return web.Response(text="[max-rows] must be > 0", status=400)
    if params['decimation-level'] is not None and params['decimation-level'] <= 0:
        return web.Response(text="[decimation-level] must be > 0", status=400)

    # --- Binary Streaming Handler ---
    resp = None

    async def stream_data(data: np.ndarray, layout, factor):
        nonlocal resp

        if resp is None:
            resp = web.StreamResponse(status=200,
                                      headers={'joule-layout': layout,
                                               'joule-decimation': str(factor)})
            resp.enable_chunked_encoding()
            await resp.prepare(request)
        await resp.write(data.tobytes())

    # --- JSON Handler ---

    data_blocks = []  # array of data segments
    data_segment = None
    decimation_factor = 1

    async def retrieve_data(data: np.ndarray, layout, factor):
        nonlocal data_blocks, data_segment, decimation_factor
        decimation_factor = factor
        if np.array_equal(data, pipes.interval_token(layout)):
            if data_segment is not None:
                data_blocks.append(data_segment.tolist())
                data_segment = None
        else:
            data = np.c_[data['timestamp'][:, None], data['data']]

            if data_segment is None:
                data_segment = data
            else:
                data_segment = np.vstack((data_segment, data))

    if json:
        callback = retrieve_data
    else:
        callback = stream_data

    # create an extraction task
    try:
        await data_store.extract(stream, params['start'], params['end'],
                                 callback=callback,
                                 max_rows=params['max-rows'],
                                 decimation_level=params['decimation-level'])
    except InsufficientDecimationError as e:
        return web.Response(text="decimated data is not available: %s" % e, status=400)
    except DataError as e:
        msg = str(e)
        if 'no such stream' in msg.lower() and (params['decimation-level'] is not None):  # pragma: no cover
            # clean up error message when user requested a particular decimation level
            msg = "requested decimation level [%d] does not exist" % params['decimation-level']
        return web.Response(text="read error: %s" % msg, status=400)

    if json:
        # put the last data_segment on
        if data_segment is not None:
            data_blocks.append(data_segment.tolist())
        return web.json_response({"data": data_blocks, "decimation_factor": decimation_factor})
    else:
        if resp is None:
            return web.json_response(text="this stream has no data", status=400)
        return resp


async def _subscribe(request: web.Request, json: bool):
    db: Session = request.app["db"]
    supervisor: Supervisor = request.app['supervisor']
    if json:
        return web.Response(text="JSON subscription not implemented", status=400)

    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    pipe = pipes.LocalPipe(stream.layout)
    try:
        unsubscribe = supervisor.subscribe(stream, pipe)
    except SubscriptionError:
        return web.Response(text="stream is not being produced", status=400)
    resp = web.StreamResponse(status=200,
                              headers={'joule-layout': stream.layout,
                                       'joule-decimation': '1'})
    resp.enable_chunked_encoding()

    try:
        await resp.prepare(request)
    except ConnectionResetError:
        unsubscribe()
        return

    try:
        while True:
            try:
                data = await pipe.read()
            except pipes.EmptyPipe:
                unsubscribe()
                return resp
            pipe.consume(len(data))
            if len(data) > 0:
                await resp.write(data.tobytes())
            if pipe.end_of_interval:
                await resp.write(pipes.interval_token(stream.layout).tobytes())
    except asyncio.CancelledError as e:
        unsubscribe()
        # propogate the CancelledError up
        raise e
    except ConnectionResetError:
        unsubscribe()


async def intervals(request: web.Request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    # parse time bounds if specified
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        else:
            start = None
        if 'end' in request.query:
            end = int(request.query['end'])
        else:
            end = None
    except ValueError:
        return web.Response(text="[start] and [end] must be an integers", status=400)

    # make sure parameters make sense
    if (start is not None and end is not None) and start >= end:
        return web.Response(text="[start] must be < [end]", status=400)

    return web.json_response(await data_store.intervals(stream, start, end))


async def write(request: web.Request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    # spawn in inserter task
    stream.is_destination = True
    db.commit()

    pipe = pipes.InputPipe(name="inbound", stream=stream, reader=request.content)

    try:
        task = await data_store.spawn_inserter(stream, pipe, insert_period=0)
        await task
    except DataError as e:
        return web.Response(text=str(e), status=400)
    except asyncio.CancelledError as e:
        raise e
    finally:
        stream.is_destination = False
        db.commit()
    return web.Response(text="ok")


async def remove(request: web.Request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse time bounds
    start = None
    end = None
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        if 'end' in request.query:
            end = int(request.query['end'])
    except ValueError:
        return web.Response(text="[start] and [end] must be integers", status=400)

    # make sure bounds make sense
    if ((start is not None and end is not None) and
            (start >= end)):
        return web.Response(text="[start] must be < [end]", status=400)

    await data_store.remove(stream, start, end)
    return web.Response(text="ok")


async def consolidate(request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse time bounds
    start = None
    end = None
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        if 'end' in request.query:
            end = int(request.query['end'])
    except ValueError:
        return web.Response(text="[start] and [end] must be integers", status=400)

    # make sure bounds make sense
    if ((start is not None and end is not None) and
            (start >= end)):
        return web.Response(text="[start] must be < [end]", status=400)

    # parse the max_gap parameter
    if 'max_gap' not in request.query:
        return web.Response(text="specify max_gap as us integer", status=400)
    try:
        max_gap = int(request.query['max_gap'])
        if max_gap <= 0:
            raise ValueError()
    except ValueError:
        return web.Response(text="max_gap must be postive integer", status=400)
    num_removed = await data_store.consolidate(stream, start, end, max_gap)
    return web.json_response(data={"num_consolidated": num_removed})




from ..reader_module import ReaderModule
from joule import utilities
import asyncio
import numpy as np
import textwrap

ARGS_DESC = """
---
:name:
  Random Reader
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:description:
  generate a random data stream
:usage:
    This is a module that generates random numbers.
    Specify width (number of elements) and the rate in Hz.
    
    | Arguments | Description        |
    |-----------|--------------------|
    |``width``      | number of elements |
    |``rate``       | data rate in Hz    |

:inputs:
  None

:outputs:
  output
  :  float32 with N elements specified by [width] argument

:stream_configs:
  #output#
     [Main]
     name = Random Data
     path = /path/to/output
     datatype = float32
     keep = 1w

     # [width] number of elements
     [Element1]
     name = Random Set 1

     [Element2]
     name = Random Set 2

     #additional elements...

:module_config:
    [Main]
    name = Random Reader
    exec_cmd = joule modules random-reader

    [Arguments]
    # number of elements
    width = 4
    # data rate in Hz
    rate = 10 

    [Outputs]
    output = /path/to/output
---
"""

OUTPUT_RATE = 1  # run in 1 second blocks


class RandomReader(ReaderModule):
    """Generate a random stream of data"""

    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")
        grp.add_argument("--width", type=int,
                         required=True,
                         help="number of elements in output")
        grp.add_argument("--rate", type=float,
                         required=True,
                         help="rate in Hz")
        parser.description = textwrap.dedent(ARGS_DESC)

    async def run(self, parsed_args, output):
        # produce output four times per second
        # figure out how much output will be in each block
        rate = parsed_args.rate
        width = parsed_args.width
        data_ts = utilities.time_now()
        data_ts_inc = 1/rate*1e6
        wait_time = 1/OUTPUT_RATE
        BLOCK_SIZE = rate/OUTPUT_RATE
        fraction_remaining = 0
        i = 0

        # print("Starting random stream: %d elements @ %0.1fHz" % (width, rate))
        while not self.stop_requested:
            float_block_size = BLOCK_SIZE+fraction_remaining
            int_block_size = int(np.floor(float_block_size))
            fraction_remaining = float_block_size - int_block_size
            data = np.random.rand(int_block_size, width)
            self.avg = np.average(data)
            self.stddev = np.std(data)
            top_ts = data_ts + int_block_size*data_ts_inc
            ts = np.array(np.linspace(data_ts, top_ts,
                                      int_block_size, endpoint=False),
                          dtype=np.uint64)
            data_ts = top_ts
            await output.write(np.hstack((ts[:, None], data)))
            await asyncio.sleep(wait_time)
            i += 1


def main():  # pragma: no cover
    r = RandomReader()
    r.start()


if __name__ == "__main__":
    main()



from joule.client.helpers.args import (module_args,
                                       load_args_from_file,
                                       validate_time_bounds,
                                       read_module_config,
                                       read_stream_configs)
from joule.client.helpers.pipes import (build_fd_pipes,
                                        build_network_pipes)



import numpy as np
import io
from struct import pack
import datetime
import asyncpg
from typing import List, Optional, Tuple
import logging

from joule.errors import DataError
from joule.models.data_stream import DataStream
import joule.utilities

log = logging.getLogger('joule')

postgres_ts_offset = 946684800000000  # January 1 2000 GMT


def data_to_bytes(data: np.ndarray) -> io.BytesIO:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = data.dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = data['data'].dtype.alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy = np.empty(data.shape, pgcopy_dtype)
    pgcopy['num_fields'] = n_elem + 1
    pgcopy['time_length'] = 8
    pgcopy['time'] = data['timestamp'] - postgres_ts_offset
    for i in range(n_elem):
        pgcopy['elem%d_length' % i] = elem_length
        pgcopy['elem%d' % i] = data['data'][:, i]
    cpy = io.BytesIO()
    # signature, flag field and and header extension (both empty)
    cpy.write(pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0))
    cpy.write(pgcopy.tobytes())
    cpy.write(pack('!h', -1))
    cpy.seek(0)
    return cpy


def bytes_to_data(buffer: io.BytesIO, dtype: np.dtype) -> np.ndarray:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = dtype['data'].alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy_dtype = np.dtype(pgcopy_dtype)
    nbytes = buffer.seek(0, io.SEEK_END)
    buffer.seek(0)
    # check the header
    header = pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0)
    rx_header = buffer.read(len(header))
    if rx_header != header:
        raise DataError("bad pgcopy header")
    row_size = pgcopy_dtype.itemsize
    if (nbytes - 21) % row_size != 0:
        raise DataError("invalid number of data bytes")
    nrows = (nbytes - 21) // row_size
    tuple_data = np.frombuffer(buffer.read(nbytes - 21), pgcopy_dtype)
    rx_data = np.empty(nrows, dtype)
    rx_data['timestamp'] = tuple_data['time'] + postgres_ts_offset
    for i in range(n_elem):
        rx_data['data'][:, i] = tuple_data['elem%d' % i]
    # reader the footer
    footer = pack('!h', -1)
    rx_footer = buffer.read()
    if footer != rx_footer:
        raise DataError("ERROR: invalid footer")
    return rx_data


def query_time_bounds(start, end):
    # bounds are [ --- )
    limits = []
    if start is not None:
        if type(start) is not datetime.datetime:
            start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
        limits.append("time >= '%s'" % start)
    if end is not None:
        if type(end) is not datetime.datetime:
            end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
        limits.append("time < '%s'" % end)
    if len(limits) > 0:
        return 'WHERE ' + ' AND '.join(limits)
    else:
        return ''


async def create_event_table(conn: asyncpg.Connection):
    sql = """CREATE TABLE IF NOT EXISTS data.events (
    time TIMESTAMP NOT NULL,
    end_Time TIMESTAMP,
    event_stream_id INTEGER NOT NULL,
    content JSONB
    )"""
    await conn.execute(sql)


async def create_stream_table(conn: asyncpg.Connection, stream: DataStream):
    n_elems = len(stream.elements)
    # create the main table
    col_type = get_psql_type(stream.datatype)
    cols = ["elem%d %s NOT NULL" % (x, col_type) for x in range(n_elems)]
    sql = "CREATE TABLE IF NOT EXISTS data.stream%d (" % stream.id + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + ");"
    await conn.execute(sql)
    sql = "SELECT create_hypertable('data.stream%d', 'time', if_not_exists=>true, chunk_target_size => 'estimate');" % stream.id
    await conn.execute(sql)

    # create interval table
    sql = "CREATE TABLE IF NOT EXISTS data.stream%d_intervals (" % stream.id + \
          "time TIMESTAMP NOT NULL);"
    await conn.execute(sql)


async def create_decimation_table(conn: asyncpg.Connection, stream: DataStream, level: int):
    n_elems = len(stream.elements)
    table_name = 'data.stream%d_%d' % (stream.id, level)
    # create decimation table (just a template)
    mean_cols = ["elem%d REAL NOT NULL" % x for x in range(n_elems)]
    min_cols = ["elem%d_min REAL NOT NULL" % x for x in range(n_elems)]
    max_cols = ["elem%d_max REAL NOT NULL" % x for x in range(n_elems)]
    cols = mean_cols + min_cols + max_cols
    sql = "CREATE TABLE IF NOT EXISTS %s (" % table_name + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + ");"
    await conn.execute(sql)
    sql = "SELECT create_hypertable('%s', 'time', if_not_exists=>true, chunk_target_size => 'estimate');" % table_name
    await conn.execute(sql)


def get_psql_type(x: DataStream.DATATYPE):
    if x == DataStream.DATATYPE.FLOAT32:
        return 'real'
    elif x == DataStream.DATATYPE.FLOAT64:
        return 'double precision'
    elif x == DataStream.DATATYPE.INT16:
        return 'smallint'
    elif x == DataStream.DATATYPE.INT32:
        return 'integer'
    elif x == DataStream.DATATYPE.INT64:
        return 'bigint'
    else:
        raise DataError("Invalid type [%r] for timescale backend" % x)


async def get_row_count(conn: asyncpg.Connection, stream: DataStream,
                        start=None, end=None):
    # hyper table approximate row count is not sensitive to data removal
    # always use the custom function
    try:
        bounds = await convert_time_bounds(conn, stream, start, end)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    if bounds is None:
        return 0  # no data
    start, end = bounds
    query = "SELECT stream_row_count(%d, '%s', '%s')" % (stream.id, start, end)
    try:
        nrows = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    return nrows


async def close_interval(conn: asyncpg.Connection, stream: DataStream, ts: int):
    # place a boundary 1us *after* ts
    base_table = "data.stream%d" % stream.id
    interval_table = "data.stream%d_intervals" % stream.id
    ts = datetime.datetime.fromtimestamp(ts / 1e6, tz=datetime.timezone.utc)
    # find the most recent data before this boundary (ts)
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (base_table, ts)
    try:
        last_ts = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        # no data tables so no need for an interval boundary
        return
    if last_ts is None:
        # no data exists before ts so no need for an interval boundary
        return
    # check if this interval is necessary
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (interval_table, ts)
    last_interval = await conn.fetchval(query)
    if last_interval is None or last_ts > last_interval:
        query = "INSERT INTO %s(time) VALUES ($1)" % interval_table
        await conn.execute(query, last_ts + datetime.timedelta(microseconds=1))


async def get_table_names(conn: asyncpg.Connection, stream: DataStream, with_schema=True) -> List[str]:
    query = r'''select table_name from information_schema.tables 
               where table_schema='data' 
               and table_type='BASE TABLE' 
               and table_name like 'stream%d\_%%';''' % stream.id
    records = await conn.fetch(query)
    if with_schema:
        return ['data.' + r['table_name'] for r in records] + ['data.stream%d' % stream.id]
    else:
        return [r['table_name'] for r in records] + ['stream%d' % stream.id]


async def get_all_table_names(conn: asyncpg.Connection, with_schema=True) -> List[str]:
    query = r'''select table_name from information_schema.tables 
               where table_schema='data' 
               and table_type='BASE TABLE' 
               and table_name like 'stream%';'''
    records = await conn.fetch(query)
    if with_schema:
        return ['data.' + r['table_name'] for r in records]
    else:
        return [r['table_name'] for r in records]


async def get_boundaries(conn: asyncpg.Connection, stream: DataStream,
                         start: Optional[int], end: Optional[int]) -> List[datetime.datetime]:
    """
    Return a list of data boundaries including the start and end of the data
    """
    bounds = await convert_time_bounds(conn, stream, start, end)
    if bounds is None:
        return []  # no data so no boundaries
    start, end = bounds
    query = "SELECT time FROM data.stream%d_intervals " % stream.id
    query += query_time_bounds(start, end)
    query += " ORDER BY time ASC"
    try:
        records = await conn.fetch(query)
    except asyncpg.UndefinedTableError:
        return []  # no data tables so no boundaries
    ts = [start] + [r['time'] for r in records] + [end]
    return [x.replace(tzinfo=datetime.timezone.utc) for x in ts]


async def convert_time_bounds(conn: asyncpg.Connection,
                              stream: DataStream,
                              start: Optional[int], end: Optional[int]) -> Optional[
    Tuple[datetime.datetime, datetime.datetime]]:
    """
    Convert Unix us timestamps to datetime objects and populate [None] values with the
    start or end of the data respectively
    """
    x = start
    y = end
    if start is None:
        query = "SELECT time FROM data.stream%d ORDER BY time ASC LIMIT 1" % stream.id
        try:
            start = await conn.fetchval(query)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if start is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        start = start.replace(tzinfo=datetime.timezone.utc)

    else:
        start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
    if end is None:
        query = "SELECT time FROM data.stream%d ORDER BY time DESC LIMIT 1" % stream.id
        try:
            end = await conn.fetchval(query) + datetime.timedelta(microseconds=1)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if end is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        end = end.replace(tzinfo=datetime.timezone.utc)
    else:
        end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
    return start, end


async def limit_time_bounds(conn: asyncpg.Connection,
                            stream: DataStream,
                            start: Optional[int], end: Optional[int]) -> Tuple[Optional[int], Optional[int]]:
    data_bounds = await convert_time_bounds(conn, stream, None, None)
    if data_bounds is None:
        return None, None
    data_start = joule.utilities.datetime_to_timestamp(data_bounds[0])
    data_end = joule.utilities.datetime_to_timestamp(data_bounds[1])
    if start is None:
        start = data_start
    else:
        start = max(start, data_start)
    if end is None:
        end = data_end
    else:
        end = min(end, data_end)
    return start, end



import click
from joule.cli.config import pass_config
from joule import api


@click.command(name="default")
@click.argument("name")
@pass_config
def node_default(config, name):
    """Change the default node."""
    try:
        api.set_default_node(name)
        click.echo("Set [%s] as the default node" % name)
    except ValueError as e:
        raise click.ClickException(str(e))



from joule.client import FilterModule
import textwrap
import numpy as np
from typing import Dict, List
from joule import Pipe
from joule.errors import ApiError, EmptyPipeError
from scipy.interpolate import interp1d
from icecream import ic

ARGS_DESC = """
---
:name:
  Merge Filter
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:description:
  combine multiple input streams into a single output stream
:usage:
  Specify a primary stream and one or more additional streams.
  The output will have timestamps from the primary stream and linearly
  interpolated data from all the additional streams
  
  
    | Arguments | Description        |
    |-----------|--------------------|
    |``primary`` | name of primary stream |

  ```
:inputs:
  primary
  : N elements
  additional
  : M elements

:outputs:
  output
  :  float32 with N+M elements auto detected from input source

:stream_configs:
  #output#
     [Main]
     name = Merged
     path = /path/to/output
     datatype = float32
     keep = 1w

     # [width] number of elements of all input streams
     [Element1]
     name = Data Column 1

     [Element2]
     name = Data Column 2

     #additional elements...

:module_config:
    [Main]
    name = Merge Filter
    exec_cmd = joule-merge-filter

    [Arguments]
    primary = input1 

    [Inputs]
    input1 = /path/to/input1
    input2 = /path/to/input2
    #... additional inputs as desired
    
    [Outputs]
    output = /path/to/output
---
"""


class MergeFilter(FilterModule):
    """Merge input streams"""

    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")
        grp.add_argument("--primary", type=str, required=True,
                         help="name of primary input stream")
        parser.description = textwrap.dedent(ARGS_DESC)

    async def run(self, parsed_args, inputs: Dict[str, Pipe], outputs):
        primary_name = parsed_args.primary

        # ---- VERIFY PARAMETERS and STREAM FORMATS -----
        if primary_name not in inputs:
            raise Exception("the specified primary [%s] is not in the inputs" % parsed_args.primary)
        primary = inputs[primary_name]
        secondaries = [inputs[name] for name in inputs.keys() if name != primary_name]
        if len(outputs) != 1:
            raise Exception("must specify one output stream")
        output = list(outputs.values())[0]
        input_width = np.sum([secondary.width for secondary in secondaries]) + primary.width
        if input_width != output.width:
            raise Exception("output must have %d elements" % input_width)
        # ---- END VERIFY ------

        if not await align_streams(primary, secondaries):
            await output.close()  # no data overlap so nothing can be processed
            return
        while not self.stop_requested:
            try:
                primary_data = await primary.read()
                # allocate the output array (we won't have any *more* data than this)
                output_data = np.empty(len(primary_data), dtype=output.dtype)
                # put in the primary data and the timestamps
                output_data['timestamp'] = primary_data['timestamp']
                output_data['data'][:, :primary.width] = make2d(primary_data['data'])
                valid_rows = len(output_data)
                offset = primary.width
                for secondary in secondaries:
                    valid_rows = min(await get_data(secondary, output_data, offset),
                                     valid_rows)
                    offset += secondary.width
                # consume the right amount from each secondary based on the last valid timestamp
                for secondary in secondaries:
                    await consume_data(secondary, output_data['timestamp'][valid_rows - 1])

                # write the output
                await output.write(output_data[:valid_rows])

                # consume the used data and check for interval breaks
                interval_break = False
                primary.consume(valid_rows)

                # stay within the same interval until all streams catch up
                if primary.end_of_interval and len(primary_data) > valid_rows:
                    primary.reread_last()
                if primary.end_of_interval:
                    interval_break = True
                for secondary in secondaries:
                    if secondary.end_of_interval:
                        # stay within the same interval until all streams catch up
                        secondary.reread_last()
                if interval_break:
                    await output.close_interval()
                    if not await align_streams(primary, secondaries):
                        break
            except EmptyPipeError:
                break
            except Exception as e:
                await output.close()
                raise e
        await output.close()


def make2d(arr: np.ndarray) -> np.ndarray:
    if len(arr.shape) == 1:
        return arr[:, None]
    else:
        return arr


async def align_streams(primary: Pipe, secondaries: List[Pipe]):
    # we need a previous sample of every secondary
    try:
        secondary_data = [await secondary.read() for secondary in secondaries]
        primary_data = await primary.read()
        # get the maximum secondary start time (we can't interpolate before this)
        try:
            latest_secondary_start = np.max([data['timestamp'][0] for data in secondary_data])
        except IndexError as e:
            breakpoint()
            print(e)
        # make sure all streams have some data larger than the latest_secondary_start
        while primary_data['timestamp'][-1] < latest_secondary_start:
            primary.consume(len(primary_data))
            primary_data = await primary.read()
        for i in range(len(secondaries)):
            while secondary_data[i]['timestamp'][-1] < latest_secondary_start:
                secondaries[i].consume(len(secondary_data[i]))
                secondary_data[i] = await secondaries[i].read()
        # find the closest ts in primary that is larger than latest_secondary_start
        start_ts = primary_data['timestamp'][primary_data['timestamp'] >= latest_secondary_start][0]
        i = 0
        for data in secondary_data:
            # find the closest ts in secondaries that is smaller than latest_secondary_start
            idx = np.argmax(data['timestamp'][data['timestamp'] <= start_ts])
            # consume up to this point
            secondaries[i].consume(idx)
            i += 1
        # consume the primary before start_ts
        too_early = primary_data['timestamp'][primary_data['timestamp'] < start_ts]
        primary.consume(len(too_early))
        # to preserve any interval breaks, reread the same data again
        primary.reread_last()
        for secondary in secondaries:
            secondary.reread_last()
    except ApiError:
        return False
    return True


async def get_data(source_pipe: Pipe, dest: np.ndarray, offset: int) -> float:
    source = await source_pipe.read()
    ts = dest['timestamp']
    max_ts = source['timestamp'][-1]
    f = interp1d(source['timestamp'], source['data'], axis=0, fill_value='extrapolate')
    resampled_data = f(ts[ts <= max_ts])
    if len(resampled_data) == 0:
        return 0  # nothing to do
    dest['data'][:len(resampled_data), offset:offset + source_pipe.width] = make2d(resampled_data)
    last_valid_ts = ts[len(resampled_data) - 1]
    dest_rows = np.argwhere(ts == last_valid_ts)[0][0] + 1
    return dest_rows


async def consume_data(source_pipe: Pipe, ts: int) -> None:
    # return
    source_pipe.reread_last()
    source = await source_pipe.read()
    used_rows = len(source['timestamp'][source['timestamp'] <= ts])
    source_pipe.consume(used_rows)
    # stay within the same interval until all streams catch up
    if source_pipe.end_of_interval and used_rows < len(source):
        source_pipe.reread_last()
        source_pipe.interval_break = False


def main():  # pragma: no cover
    r = MergeFilter()
    r.start()


if __name__ == "__main__":
    main()



from joule import errors


class BaseSession:

    def __init__(self):
        self.url = ""
        self._session = None
        self.ssl_context = None
        self.cafile = ""

    async def get_session(self):
        return self._session

    async def get(self, path, params=None):
        return await self._request("GET", path, params=params)

    async def post(self, path, json=None, params=None, data=None, chunked=None):
        return await self._request("POST", path, json=json,
                                   params=params, data=data, chunked=chunked)

    async def put(self, path, json):
        return await self._request("PUT", path, json=json)

    async def delete(self, path, params):
        return await self._request("DELETE", path, params=params)

    async def _request(self, method, path, data=None, json=None, params=None, chunked=None):
        raise errors.ApiError("Implement in child class")

    async def close(self):
        if self._session is not None:
            await self._session.close()
        self._session = None



from typing import Optional, Union, Dict, List

from .session import BaseSession
from .folder_type import Folder

from joule import errors


class DataStream:
    """
    API DataStream model. See :ref:`sec-node-data-stream-actions` for details on using the API to
    manipulate data streams. Streams are locked if they are active or statically configured. When
    creating a stream manually, omit the ID and status attributes (**is_\***, **active**, and **locked**),
    these are set by the Joule server.

    Parameters:
        name (str): stream name, must be unique in the parent
        description (str): optional field
        datatype (str): element datatype
        keep_us (int): store the last N microseconds of data (-1 to keep all and 0 to keep none)
        is_configured: is the stream statically configured with a \*.conf file
        is_source: is the stream an active data source
        is_destination: is the stream an active data destination
        active (bool): is the stream a source or destination
        locked (bool): is the stream active or configured
        decimate (bool): is the stream data decimated for visualization
        elements (List[Element]): list of the stream elements

    """

    def __init__(self,
                 name: str = "", description: str = "",
                 datatype: str = "float32", keep_us: int = -1,
                 elements: List['Element'] = None):
        self._id = None
        self.name = name
        self.description = description
        self.datatype = datatype
        self.keep_us = keep_us  # -1 = KEEP ALL
        self.is_configured = False
        self.is_source = False
        self.is_destination = False
        self.locked = False
        self.active = False
        self.decimate = True

        if elements is None:
            elements = []
        self.elements = elements

    def __repr__(self):
        return "<joule.api.DataStream id=%r name=%r description=%r datatype=%r is_configured=%r is_source=%r is_destination=%r locked=%r decimate=%r>" % (
            self._id, self.name, self.description,
            self.datatype, self.is_configured,
            self.is_source, self.is_destination,
            self.locked, self.decimate)

    @property
    def id(self) -> int:
        if self._id is None:
            raise errors.ApiError("this is a local model with no ID. See API docs")
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    @property
    def layout(self):
        return self.datatype.lower() + '_' + str(len(self.elements))

    def to_json(self) -> Dict:
        return {
            "id": self._id,
            "name": self.name,
            "description": self.description,
            "is_configured": self.is_configured,
            "is_source": self.is_source,
            "is_destination": self.is_destination,
            "datatype": self.datatype,
            "keep_us": self.keep_us,
            "locked": self.locked,
            "active": self.active,
            "decimate": self.decimate,
            "elements": [e.to_json() for e in self.elements]
        }


def from_json(json) -> DataStream:
    my_stream = DataStream()
    my_stream.id = json['id']
    my_stream.name = json['name']
    my_stream.description = json['description']
    my_stream.datatype = json['datatype']
    my_stream.decimate = json['decimate']
    my_stream.keep_us = json['keep_us']
    my_stream.is_configured = json['is_configured']
    my_stream.is_source = json['is_source']
    my_stream.is_destination = json['is_destination']
    my_stream.locked = json['locked']
    my_stream.active = json['active']
    my_stream.elements = [elem_from_json(item) for item in json['elements']]
    return my_stream


class Element:
    """
    API Element model. Streams have one or more elements. See :ref:`sec-streams` for details on the stream data model.

    Parameters:
        id (int): unique numeric ID assigned by Joule
        index (int): column position in the data array (0 = first element)
        name (str): element name
        units (str): unit of measurement, may be any string
        plottable (bool): should the element be visible in the Lumen plotting interface
        display_type [continous|discrete|event]: plot type, defaults to continuous
        offset (float): offset data visualization by ``y=(x-offset)*scale_factor``
        scale_factor (float): scale data visualation with above equation
        default_max (float): fix auto scale max (set to None to fit plotted data)
        default_min (float): fix auto scale min (set to None to fit plotted data)
    """

    def __init__(self, name: str = "", units: str = "",
                 plottable: bool = True,
                 display_type: str = 'continuous'):
        self.id = None
        self.index = None
        self.name = name
        self.units = units
        self.plottable = plottable
        self.display_type = display_type
        self.offset = 0
        self.scale_factor = 1.0
        self.default_max = None
        self.default_min = None

    def __repr__(self):
        return "<joule.api.Element id=%r index=%r, name=%r units=%r plottable=%r display_type=%r>" % (
            self.id, self.index, self.name,
            self.units, self.plottable,
            self.display_type)

    def to_json(self) -> Dict:
        return {
            'id': self.id,
            'index': self.index,
            'name': self.name,
            'units': self.units,
            'plottable': self.plottable,
            'display_type': self.display_type,
            'offset': self.offset,
            'scale_factor': self.scale_factor,
            'default_max': self.default_max,
            'default_min': self.default_min
        }


def elem_from_json(json) -> Element:
    my_elem = Element()
    my_elem.id = json['id']
    my_elem.index = json['index']
    my_elem.name = json['name']
    my_elem.units = json['units']
    my_elem.plottable = json['plottable']
    my_elem.display_type = json['display_type']
    my_elem.offset = json['offset']
    my_elem.scale_factor = json['scale_factor']
    my_elem.default_min = json['default_min']
    my_elem.default_max = json['default_max']
    return my_elem


class DataStreamInfo:
    """
        API DataStreamInfo model. Received from :meth:`Node.data_stream_info` and should not be created directly.

        .. warning::
            Rows and Bytes values are approximate

        Parameters:
            start (int): timestamp in UNIX microseconds of the first data element
            end (int): timestamp in UNIX microsseconds of the last data element
            rows (int): approximate rows of data in the stream
            bytes (int): approximate size of the data on disk
            total_time (int): data duration in microseconds (start-end)

        """

    def __init__(self, start: Optional[int], end: Optional[int], rows: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.rows = rows
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<joule.api.DataStreamInfo start=%r end=%r rows=%r, total_time=%r>" % (
            self.start, self.end, self.rows, self.total_time)


def info_from_json(json) -> DataStreamInfo:
    if json is not None:

        return DataStreamInfo(json['start'],
                              json['end'],
                              json['rows'],
                              json['total_time'],
                              json['bytes'])
    else:
        return DataStreamInfo(None,
                              None,
                              0,
                              0,
                              0)


async def data_stream_delete(session: BaseSession,
                             stream: Union[DataStream, str, int]):
    data = {}
    if type(stream) is DataStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")

    await session.delete("/stream.json", data)


async def data_stream_create(session: BaseSession,
                             stream: DataStream, folder: Union[Folder, str, int]) -> DataStream:
    data = {"stream": stream.to_json()}

    if type(folder) is Folder:
        data["dest_id"] = folder.id
    elif type(folder) is int:
        data["dest_id"] = folder
    elif type(folder) is str:
        data["dest_path"] = folder
    else:
        raise errors.ApiError("Invalid folder datatype. Must be Folder, Path, or ID")

    resp = await session.post("/stream.json", json=data)
    return from_json(resp)


async def data_stream_info(session: BaseSession,
                           stream: Union[DataStream, str, int]) -> DataStreamInfo:
    data = {}

    if type(stream) is DataStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")

    resp = await session.get("/stream.json", data)
    return info_from_json(resp['data_info'])


async def data_stream_get(session: BaseSession,
                          stream: Union[DataStream, str, int]) -> DataStream:
    data = {}

    if type(stream) is DataStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")

    resp = await session.get("/stream.json", data)
    return from_json(resp)


async def data_stream_update(session: BaseSession,
                             stream: DataStream) -> None:
    await session.put("/stream.json", {"id": stream.id,
                                       "stream": stream.to_json()})


async def data_stream_move(session: BaseSession,
                           source: Union[DataStream, str, int],
                           destination: Union[Folder, str, int]) -> None:
    data = {}

    if type(source) is DataStream:
        data["src_id"] = source.id
    elif type(source) is int:
        data["src_id"] = source
    elif type(source) is str:
        data["src_path"] = source
    else:
        raise errors.ApiError("Invalid source datatype. Must be DataStream, Path, or ID")

    if type(destination) is Folder:
        data["dest_id"] = destination.id
    elif type(destination) is int:
        data["dest_id"] = destination
    elif type(destination) is str:
        data["dest_path"] = destination
    else:
        raise errors.ApiError("Invalid destination datatype. Must be Folder, Path, or ID")
    await session.put("/stream/move.json", data)


async def data_stream_annotation_delete(session: BaseSession,
                                        stream: Union[DataStream, str, int],
                                        start: Optional[int] = None,
                                        end: Optional[int] = None):
    data = {}
    if start is not None:
        data["start"] = start
    if end is not None:
        data["end"] = end

    if type(stream) is DataStream:
        data["stream_id"] = stream.id
    elif type(stream) is int:
        data["stream_id"] = stream
    elif type(stream) is str:
        data["stream_path"] = stream
    else:
        raise errors.ApiError("Invalid source datatype. Must be DataStream, Path, or ID")

    await session.delete("/stream/annotations.json", params=data)



from aiohttp import web
from sqlalchemy.exc import CircularDependencyError
from sqlalchemy.orm import Session
from typing import List

from joule.models import folder, Folder, EventStore, DataStore, DataStream, EventStream
from joule.errors import ConfigurationError


async def index(request: web.Request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    event_store: DataStore = request.app["event-store"]

    root = folder.root(db)
    data_streams = db.query(DataStream).all()
    data_streams_info = await data_store.info(data_streams)
    event_streams = db.query(EventStream).all()
    event_streams_info = await event_store.info(event_streams)
    resp = root.to_json(data_streams_info, event_streams_info)
    return web.json_response(resp)


async def info(request: web.Request):
    db: Session = request.app["db"]
    if 'path' in request.query:
        my_folder = folder.find(request.query['path'], db)
    elif 'id' in request.query:
        my_folder = db.query(Folder).get(request.query["id"])
    else:
        return web.Response(text="specify an id or path", status=400)
    if my_folder is None:
        return web.Response(text="folder does not exist", status=404)
    return web.json_response(my_folder.to_json())


async def move(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()
    # find the folder
    if 'src_path' in body:
        my_folder = folder.find(body['src_path'], db)
    elif 'src_id' in body:
        my_folder = db.query(folder.Folder).get(body["src_id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_folder is None:
        return web.Response(text="folder does not exist", status=404)
    # make sure the folder is not locked
    if my_folder.locked:
        return web.Response(text="folder is locked", status=400)
    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            return web.Response(text="Destination error: %s" % str(e), status=400)
    elif 'dest_id' in body:
        destination = db.query(Folder).get(body["dest_id"])
    else:
        return web.Response(text="specify a destination", status=400)
    # make sure name is unique in the folder
    peer_names = [f.name for f in destination.children]
    if my_folder.name in peer_names:
        db.rollback()
        return web.Response(text="a folder with this name is in the destination folder",
                            status=400)
    try:
        # make sure this does not create a circular graph
        parent = destination.parent
        while parent is not None:
            if parent == my_folder:
                raise ValueError()
            parent = parent.parent
        destination.children.append(my_folder)
        db.commit()
    except (CircularDependencyError, ValueError):
        db.rollback()
        return web.Response(text="cannot place a parent in a child folder",
                            status=400)
    return web.json_response({"stream": my_folder.to_json()})


async def update(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()
    if 'id' not in body:
        return web.Response(text="Invalid request: specify id", status=400)

    my_folder: Folder = db.query(Folder).get(body['id'])
    if my_folder is None:
        return web.Response(text="folder does not exist", status=404)
    if my_folder.locked:
        return web.Response(text="folder is locked", status=400)
    if 'folder' not in body:
        return web.Response(text="Invalid request: specify folder as JSON", status=400)
    try:
        attrs = dict(body['folder'])
    except ValueError:
        return web.Response(text="error: [folder] attribute must be JSON", status=400)
    try:
        my_folder.update_attributes(attrs)
        # make sure name is unique in the folder
        peer_names = [f.name for f in my_folder.parent.children if f.id != my_folder.id]
        if my_folder.name in peer_names:
            db.rollback()
            return web.Response(text="a folder with this name is in the destination folder",
                                status=400)
        db.commit()
    except (ValueError, ConfigurationError) as e:
        db.rollback()
        return web.Response(text="error: %s" % e, status=400)
    return web.json_response(data=my_folder.to_json())


async def delete(request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    event_store: EventStore = request.app["event-store"]
    # find the requested folder
    if 'path' in request.query:
        my_folder: Folder = folder.find(request.query['path'], db)
    elif 'id' in request.query:
        my_folder = db.query(Folder).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_folder is None:
        return web.Response(text="folder does not exist", status=404)
    if my_folder.locked:
        return web.Response(text="folder is locked", status=400)
    if 'recursive' in request.query and request.query["recursive"] == '1':
        await _recursive_delete(my_folder.children, db, data_store, event_store)
    elif len(my_folder.children) > 0:
        return web.Response(text="specify [recursive] or remove child folders first", status=400)

    for stream in my_folder.data_streams:
        if stream.locked:  # pragma: no cover
            # shouldn't ever get here because of the check above, but just in case
            return web.Response(text="cannot delete folder with locked streams", status=400)
        await data_store.destroy(stream)
        db.delete(stream)
    for stream in my_folder.event_streams:
        await event_store.destroy(stream)

    db.delete(my_folder)
    db.commit()
    return web.Response(text="ok")


async def _recursive_delete(folders: List[Folder], db: Session, data_store: DataStore,
                            event_store: EventStore):
    if len(folders) == 0:
        return
    for f in folders:
        for stream in f.data_streams:
            if stream.locked:  # pragma: no cover
                # shouldn't ever get here because a locked folder shouldn't call us
                raise ConfigurationError("cannot delete folder with locked streams")
            await data_store.destroy(stream)
            db.delete(stream)
        for stream in f.event_streams:
            await event_store.destroy(stream)
            db.delete(stream)
        await _recursive_delete(f.children, db, data_store, event_store)
        db.delete(f)



from joule.client import ReaderModule
from joule import utilities
import textwrap
import numpy as np
import asyncio

ARGS_DESC = """
---
:name:
  File Reader
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:description:
  read data from file or named pipe
:usage:
  Read data in from a file or a named pipe.
  Optionally add timestamps to data (note, this should only be
  used for real time sources. The 8us difference in example below
  is the actual time taken to read the file line)

  Example:

  ```
    $> cat /tmp/file
    3 4
    4 6
    $> joule-file-reader --timestamps=yes --file=/tmp/file
    1485274825371860 3.0 4.0
    1485274825371868 5.0 6.0
  ```
:inputs:
  None

:outputs:
  output
  :  float32 with N elements auto detected from input source

:stream_configs:
  #output#
     [Main]
     name = File Data
     path = /path/to/output
     datatype = float32
     keep = 1w

     # [width] number of elements
     [Element1]
     name = Data Column 1

     [Element2]
     name = Data Column 2

     #additional elements...

:module_config:
    [Main]
    name = Random Reader
    exec_cmd = joule-file-reader

    [Arguments]
    timestamps = yes 
    file = /absolute/file/path

    [Outputs]
    output = /path/to/output
---
"""


class FileReader(ReaderModule):
    """Read data from a file"""

    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")

        grp.add_argument("file", help="file name")
        grp.add_argument("-d", "--delimiter", default=",",
                         choices=[" ", ","],
                         help="character between values")
        grp.add_argument("-t", "--timestamp", type=utilities.yesno,
                         help="apply timestamps to data")
        parser.description = textwrap.dedent(ARGS_DESC)

    async def run(self, parsed_args, output):
        with open(parsed_args.file, 'r') as f:
            for line in f:
                data = np.fromstring(line, dtype=float,
                                     sep=parsed_args.delimiter)

                if parsed_args.timestamp:
                    data = np.insert(data, 0, utilities.time_now())
                await output.write(np.array([data]))
                #TODO is this necessary?
                await asyncio.sleep(0.1)
                if self.stop_requested:
                    break


def main():  # pragma: no cover
    r = FileReader()
    r.start()


if __name__ == "__main__":
    main()



import numpy as np
import logging
import asyncio
from joule.models.pipes import Pipe, find_interval_token
from joule.models.pipes.errors import PipeError, EmptyPipe

log = logging.getLogger('joule')


class InputPipe(Pipe):

    def __init__(self, name=None, stream=None, layout=None,
                 reader=None, reader_factory=None,
                 close_cb=None, buffer_size=10000):
        super().__init__(name=name, stream=stream, direction=Pipe.DIRECTION.INPUT, layout=layout)
        self.reader_factory = reader_factory
        self.reader = reader
        self.close_cb = close_cb
        self._reader_close = None
        self.byte_buffer = b''
        self.unprocessed_np_buffer = b''
        self.interval_break = False
        self._reread = False
        self._last_read = False  # set flag to indicate the pipe is done,
                                 # the *next* read will generate an EmptyPipe exception
                                 # this prevents pipes that are partially consumed from running forever
        # tunable constant
        self.BUFFER_SIZE = buffer_size
        """Note: The StreamReader.read coroutine hangs even if the write
        side of the pipe is closed so the call is wrapped in a wait_for"""
        self.TIMEOUT_INTERVAL = 0.5
        self.buffer = np.zeros(self.BUFFER_SIZE * 2, dtype=self.dtype)
        self.last_index = 0

    async def read(self, flatten=False) -> np.ndarray:
        if self.reader is None:
            self.reader, self._reader_close = await self.reader_factory()
        if self.closed:
            # this happens if close is called before the first read
            if self._reader_close is not None:
                self._reader_close()
            raise PipeError("Cannot read from a closed pipe")
        rowbytes = self.dtype.itemsize
        max_rows = self.BUFFER_SIZE - (
                self.last_index + len(self.unprocessed_np_buffer) % rowbytes)

        if max_rows == 0:
            # buffer is full, this must be consumed before a new read
            return self._format_data(self.buffer[:self.last_index], flatten)

        # if reread is set just return the old data
        if self._reread:
            self._reread = False
            if self.last_index == 0:
                raise PipeError("No data left to reread")
            return self._format_data(self.buffer[:self.last_index], flatten)

        # make sure we get at least one full row of data from read (depends on datatype)
        raw = b''
        while True:
            new_data = b''
            if self.reader.at_eof():
                # do not raise an exception, but is_empty() will return True
                #if self._last_read:
                #    raise EmptyPipe()  # this data has already been read once
                if (len(self.unprocessed_np_buffer) == 0 and
                        self.last_index == 0):
                    raise EmptyPipe()
                if len(self.unprocessed_np_buffer) == 0:
                    # no new data is coming in, read() will just return
                    # previously viewed data
                    self._last_read = True
                break
            try:
                new_data = await asyncio.wait_for(self.reader.read(max_rows * rowbytes),
                                             self.TIMEOUT_INTERVAL)
            except asyncio.TimeoutError:
                pass
            raw += new_data
            if len(raw) < self.dtype.itemsize:
                await asyncio.sleep(0.1)
            else:
                break
        # extra_bytes: number of leftover bytes after % rowbytes
        # byte_buffer: the extra_bytes from the last read
        # unprocessed_np_buffer: data leftover from an interval break in the previous read
        extra_bytes = (len(raw) + len(self.byte_buffer)) % rowbytes

        if extra_bytes > 0:
            np_buffer = self.byte_buffer + raw[:-extra_bytes]
            self.byte_buffer = raw[-extra_bytes:]
        elif len(self.byte_buffer) > 0:
            np_buffer = self.byte_buffer + raw
            self.byte_buffer = b''
        else:  # common case where byte_buffer is empty and no extra bytes
            np_buffer = raw
            self.byte_buffer = b''

        # append unprocessed np_buffer from previous read
        if len(self.unprocessed_np_buffer) > 0:
            self.unprocessed_np_buffer = self.unprocessed_np_buffer + np_buffer
            # check if we can process all the data, if not
            # store the extra in unprocessed_np_buffer
            max_bytes = max_rows * rowbytes
            if len(self.unprocessed_np_buffer) <= max_bytes:
                np_buffer = self.unprocessed_np_buffer
                self.unprocessed_np_buffer = b''
            else:
                np_buffer = self.unprocessed_np_buffer[:max_bytes]
                self.unprocessed_np_buffer = self.unprocessed_np_buffer[max_bytes:]

        # check for an interval
        self.interval_break = False
        loc = find_interval_token(np_buffer, self.layout)
        if loc is not None:
            self.unprocessed_np_buffer = np_buffer[loc[1]:] + self.unprocessed_np_buffer
            np_buffer = np_buffer[:loc[0]]
            self.interval_break = True
        data = np.frombuffer(np_buffer, dtype=self.dtype)

        # append data onto buffer
        self.buffer[self.last_index:self.last_index + len(data)] = data

        self.last_index += len(data)
        return self._format_data(self.buffer[:self.last_index], flatten)

    def reread_last(self):
        if self.last_index == 0:
            raise PipeError("No data left to reread")
        self._reread = True

    def consume(self, num_rows):
        if num_rows == 0:
            return  # nothing to do
        if num_rows < 0:
            raise PipeError("consume called with negative offset: %d" % num_rows)
        if num_rows > self.last_index:
            raise PipeError("cannot consume %d rows: only %d available"
                            % (num_rows, self.last_index))
        self.buffer = np.roll(self.buffer, -1 * num_rows)
        self.last_index -= num_rows

    def is_empty(self):
        # 0.) if the read() has already signaled its finished the pipe is empty
        if self._last_read:
            return True
        # 1.) make sure the sender is closed
        if not self.reader.at_eof():
            return False
        # 2.) make sure there is no data left in the buffer
        if len(self.unprocessed_np_buffer) == 0 and self.last_index == 0:
            return True
        return False

    @property
    def end_of_interval(self):
        return self.interval_break

    async def close(self):
        if self._reader_close is not None:
            self._reader_close()
            self._reader_close = None
        if self.closed:  # don't execute cb more than once
            return
        self.closed = True
        if self.close_cb is not None:
            # used to close socket pipes
            await self.close_cb()
            self.close_cb = None



import asyncpg
import json

from typing import List, Optional, TYPE_CHECKING, Dict, Callable, Tuple
from joule.models.data_store import psql_helpers
import joule.utilities

if TYPE_CHECKING:
    from joule.models import EventStream


class EventStore:

    def __init__(self, dsn=""):
        self.dsn = dsn
        self.pool = None

    async def initialize(self, pool=None) -> None:
        if pool is None:
            if self.dsn == "":
                raise ValueError("Must specify either dsn or pool")
            self.pool = await asyncpg.create_pool(self.dsn, command_timeout=60)
        else:
            self.pool = pool
        async with self.pool.acquire() as conn:
            await psql_helpers.create_event_table(conn)

    async def insert(self, stream: 'EventStream', events: List):
        mapper = map(event_to_record_mapper(stream), events)
        async with self.pool.acquire() as conn:
            await conn.copy_records_to_table("events",
                                             schema_name='data',
                                             records=mapper)

    async def count(self, stream: 'EventStream',
                    start: Optional[int] = None, end: Optional[int] = None) -> int:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = "SELECT count(*) FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end)
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        query += where_clause
        async with self.pool.acquire() as conn:
            record = await conn.fetch(query)
            return record[0]['count']

    async def extract(self, stream: 'EventStream',
                      start: Optional[int] = None, end: Optional[int] = None) -> List[Dict]:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = "SELECT time, end_time, content FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end)
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        query += where_clause + " ORDER BY time ASC"
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            return list(map(record_to_event, records))

    async def remove(self, stream: 'EventStream', start: Optional[int] = None, end: Optional[int] = None):
        query = "DELETE FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end)
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        query += where_clause
        async with self.pool.acquire() as conn:
            return await conn.execute(query)

    async def destroy(self, stream: 'EventStream'):
        await self.remove(stream)

    async def info(self, streams: List['EventStream']) -> Dict[int, 'StreamInfo']:
        results = {}
        async with self.pool.acquire() as conn:
            for my_stream in streams:
                rows = await conn.fetchval("SELECT COUNT(*) FROM data.events WHERE event_stream_id=%d" % my_stream.id)
                query = "SELECT time FROM data.events WHERE event_stream_id=%d ORDER BY time ASC LIMIT 1" % my_stream.id
                start_time = await conn.fetchval(query)
                if start_time is not None:
                    start_time = joule.utilities.datetime_to_timestamp(start_time)
                query = "SELECT time FROM data.events WHERE event_stream_id=%d ORDER BY time DESC LIMIT 1" % my_stream.id
                end_time = await conn.fetchval(query)
                if end_time is not None:
                    end_time = joule.utilities.datetime_to_timestamp(end_time)
                if start_time is not None:
                    total_time = end_time - start_time
                else:
                    total_time = 0
                total_bytes = 0  # TODO
                results[my_stream.id] = StreamInfo(start_time, end_time, rows, total_time, total_bytes)
        return results

    async def close(self):
        await self.pool.close()


class StreamInfo:
    def __init__(self, start: Optional[int], end: Optional[int], event_count: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.event_count = event_count
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<DataStreamInfo start=%r end=%r event_count=%r, total_time=%r>" % (
            self.start, self.end, self.event_count, self.total_time)

    def to_json(self):
        return {
            "start": self.start,
            "end": self.end,
            "event_count": self.event_count,
            "bytes": self.bytes,
            "total_time": self.total_time
        }


def event_to_record_mapper(stream: 'EventStream') -> Callable[[Dict], Tuple]:
    stream_id = stream.id

    def mapper(event: Dict) -> Tuple:
        start = joule.utilities.timestamp_to_datetime(event['start_time'])
        if event['end_time'] is not None:
            if event['start_time'] >= event['end_time']:
                raise ValueError("Event end [%d] cannot be before start [%d]" % (
                    event['start_time'], event['end_time']
                ))
            end = joule.utilities.timestamp_to_datetime(event['end_time'])
        else:
            end = None
        return start, end, stream_id, json.dumps(event['content'])

    return mapper


def record_to_event(record: asyncpg.Record) -> Dict:
    if record['end_time'] is not None:
        end = joule.utilities.datetime_to_timestamp(record['end_time'])
    else:
        end = None
    return {'start_time': joule.utilities.datetime_to_timestamp(record['time']),
            'end_time': end,
            'content': json.loads(record['content'])}



import asyncio
import logging

import joule.utilities
from . import base_module
from joule import errors


class FilterModule(base_module.BaseModule):

    async def setup(self, parsed_args, app, inputs, outputs):
        """
        Configure the module, executes before :meth:`run`

        Args:
            parsed_args:
            app:
            inputs:
            outputs:

        Returns:

        """
        pass

    async def run(self, parsed_args, inputs, outputs):
        """
        This method must be implemented. It should run in a loop, if it returns the module stops.

        Args:
            parsed_args: parsed command line arguments, configure with :meth:`joule.BaseModule.custom_args`
            inputs: pipe connections to input streams indexed by name (specified in the module configuration file).
            outputs: pipe connections to output streams indexed by name (specified in the module configuration file).

        .. code-block:: python

            class ModuleDemo(FilterModule):

                def run(self, parsed_args, inputs, outputs):
                    raw = inputs["raw"]
                    filtered = outputs["filtered"]
                    # this filter just passes the input through to the output
                     while(not self.stop_requested):
                        data = await raw.read()
                        await filtered.write(data)
                        raw.consume(len(data))

                #... other module code
        """
        assert False, "implement in child class"  # pragma: no cover

    async def run_as_task(self, parsed_args, app) -> asyncio.Task:
        (input_streams, output_streams) = await self._parse_streams(parsed_args)
        missing_intervals = await self._compute_missing_intervals(input_streams, output_streams, parsed_args)
        return asyncio.create_task(self._task(missing_intervals, input_streams, output_streams, parsed_args, app))

    async def _task(self, pending_intervals, input_streams, output_streams, parsed_args, app):
        if len(pending_intervals) == 0:
            print("Nothing to do, data is already filtered")
        # if an interval is None this indicates the pipes should be live
        for interval in pending_intervals:
            if interval is not None:
                print("%s => %s" % (joule.utilities.timestamp_to_human(interval[0]),
                                    joule.utilities.timestamp_to_human(interval[1])))
            elif parsed_args.live:
                print("Running filter on live input data")
            try:
                (pipes_in, pipes_out) = await self._build_pipes_new(interval, input_streams,
                                                                    output_streams, parsed_args.pipes)
            except errors.ApiError as e:
                logging.error(str(e))
                return asyncio.create_task(asyncio.sleep(0))
            await self.setup(parsed_args, app, pipes_in, pipes_out)
            try:
                await self.run(parsed_args, pipes_in, pipes_out)
            except (asyncio.CancelledError, errors.EmptyPipeError):
                pass
            for pipe in pipes_out.values():
                await pipe.close()



from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey, types
import datetime

from joule.models.meta import Base
from joule.errors import ApiError


class Annotation(Base):
    """
    Attributes:
        title (str): title of the annotation
        content (str): a longer description if necessary
        start (date): the time of the event
        end (date): if the annotation is a range, the end time
    """
    __tablename__ = 'annotation'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    title: str = Column(String)
    content: str = Column(String)
    start: datetime = Column(types.TIMESTAMP, nullable=False)
    end: datetime = Column(types.TIMESTAMP, default=None)

    stream_id: int = Column(Integer, ForeignKey('metadata.stream.id'))
    stream: 'DataStream' = relationship("DataStream", back_populates="annotations")

    def __repr__(self):
        return "<Annotation(title='%s', content='%s', stream_id=%s)>" % (self.title, self.content, self.stream_id)

    def to_json(self):
        json = {
            'id': self.id,
            'title': self.title,
            'content': self.content,
            'start': int(self.start.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6),
            'stream_id': self.stream_id
        }
        if self.end is not None:
            json['end'] = int(self.end.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)
        else:
            json['end'] = None
        return json

    def update_attributes(self, json):
        if 'title' in json:
            self.title = json['title']
        if 'content' in json:
            self.content = json['content']


def from_json(json):
    try:
        my_annotation = Annotation(title=json['title'])
        ts = float(json['start'])
        my_annotation.start = datetime.datetime.utcfromtimestamp(ts / 1e6)
        if 'end' in json and json['end'] is not None:
            ts = float(json['end'])
            my_annotation.end = datetime.datetime.utcfromtimestamp(ts / 1e6)
        if 'content' in json:
            my_annotation.content = json['content']
    except (ValueError, KeyError, TypeError):
        raise ApiError("invalid values for annotation")
    return my_annotation



from sqlalchemy.orm import relationship
from sqlalchemy import (Column, Integer, String,
                        Boolean, ForeignKey)
from sqlalchemy.dialects.postgresql import BIGINT
from typing import TYPE_CHECKING, Dict

from joule.errors import ConfigurationError

from joule.models.meta import Base
#from joule.models.data_store.event_store import StreamInfo

if TYPE_CHECKING:
    from joule.models import (Folder)  # pragma: no cover
    from joule.models.data_store.event_store import StreamInfo

class EventStream(Base):
    """
    Attributes:
        name (str): stream name
        description (str): stream description
        decimate (bool): whether to store decimated data for event visualization
        folder (joule.Folder): parent Folder
        keep_us (int): microseconds of data to keep (KEEP_NONE=0, KEEP_ALL=-1).
    """
    __tablename__ = 'eventstream'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    name: str = Column(String, nullable=False)

    decimate: bool = Column(Boolean, default=True)

    KEEP_ALL = -1
    KEEP_NONE = 0
    keep_us: int = Column(BIGINT, default=KEEP_ALL)

    description: str = Column(String)
    folder_id: int = Column(Integer, ForeignKey('metadata.folder.id'))
    folder: "Folder" = relationship("Folder", back_populates="event_streams")

    def update_attributes(self, attrs: Dict) -> None:
        if 'name' in attrs:
            self.name = validate_name(attrs['name'])
        if 'description' in attrs:
            self.description = attrs['description']

    def to_json(self, info: Dict[int, 'StreamInfo'] = None) -> Dict:
        """
          Args:
            info: optional content added to ``data_info`` field

        Returns: Dictionary of DataStream attributes

        """

        if info is not None and self.id in info:
            data_info = info[self.id].to_json()
        else:
            data_info = None

        return {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'data_info': data_info
        }

    def __str__(self):
        return "EventStream [{name}]".format(name=self.name)

    def __repr__(self):
        return "<EventStream(id=%r, name='%s')>" % (self.id, self.name)


def from_json(data: Dict) -> EventStream:
    """
    Construct an EventStream from a dictionary of attributes produced by
    :meth:`EventStream.to_json`

    Args:
        data: attribute dictionary

    Returns: EventStream

    """

    return EventStream(id=data["id"],
                       name=validate_name(data["name"]),
                       description=data["description"])


def validate_name(name: str) -> str:
    if name is None or len(name) == 0:
        raise ConfigurationError("missing name")
    if '/' in name:
        raise ConfigurationError("invalid name, '\\' not allowed")
    return name



import asyncio
import aiohttp
from aiohttp import web
import aiohttp_jinja2
import jinja2
import os
import random
import numpy as np

from joule.client.filter_module import FilterModule

CSS_DIR = os.path.join(os.path.dirname(__file__), 'assets', 'css')
JS_DIR = os.path.join(os.path.dirname(__file__), 'assets', 'js')
TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'assets', 'templates')

ARGS_DESC = """
TODO
"""


class Visualizer(FilterModule):  # pragma: no cover

    async def setup(self, parsed_args, app, inputs, outputs):
        loader = jinja2.FileSystemLoader(TEMPLATES_DIR)
        aiohttp_jinja2.setup(app, loader=loader)
        app["title"] = parsed_args.title

        self.elements = []
        dom_id = 0  # DOM id for javascript manipulation
        for pipe in inputs.values():
            for element in pipe.stream.elements:
                self.elements.append({
                    'stream': pipe.stream.name,
                    'element': element.name,
                    'value': '&mdash;',
                    'min': '&mdash;',
                    'max': '&mdash;',
                    'id': dom_id
                })
                dom_id += 1
        if len(self.elements) == 0:
            self.mock_data = True
            self.elements = self._create_mock_elements(4)

        else:
            self.mock_data = False

    def custom_args(self, parser):
        parser.add_argument("--title", default="Data Visualizer", help="page title")
        parser.description = ARGS_DESC

    async def run(self, parsed_args, inputs, outputs):
        if self.mock_data:
            while True:
                self._update_mock_data()
                await asyncio.sleep(1)
        while not self.stop_requested:
            offset = 0
            for pipe in inputs.values():
                data = await pipe.read()
                pipe.consume(len(data))
                if len(data) == 0:
                    continue
                for i in range(len(pipe.stream.elements)):
                    data_mean = float(np.mean(data['data'][i]))
                    data_min = float(np.min(data['data'][i]))
                    data_max = float(np.max(data['data'][i]))
                    self.elements[i + offset]['value'] = data_mean
                    # compute the new min value
                    if type(self.elements[i + offset]['min']) is str:
                        self.elements[i + offset]['min'] = data_min
                    else:
                        global_min = self.elements[i + offset]['min']
                        self.elements[i + offset]['min'] = min((data_min, global_min))
                    # compute the new max value
                    if type(self.elements[i + offset]['max']) is str:
                        self.elements[i + offset]['max'] = data_max
                    else:
                        global_max = self.elements[i + offset]['max']
                        self.elements[i + offset]['max'] = max((data_max, global_max))
                offset += len(pipe.stream.elements)

            await asyncio.sleep(0.5)

    def routes(self):
        return [
            web.get('/', self.index),
            web.get('/data.json', self.data),
            web.post('/reset.json', self.reset),
            web.static('/assets/css', CSS_DIR),
            web.static('/assets/js', JS_DIR),
        ]

    @aiohttp_jinja2.template('index.jinja2')
    async def index(self, request):
        return {'title': request.app['title'], 'elements': self.elements}

    async def data(self, request):
        return web.json_response(data=self.elements)

    async def reset(self, request):

        # clear the max and min values
        for element in self.elements:
            element['min'] = "&mdash;"
            element['max'] = "&mdash;"

        return web.json_response(data=self.elements)

    def _update_mock_data(self):
        for element in self.elements:
            element['value'] = random.randint(1, 101)
            element['min'] = element['value'] - random.randint(1, 101)
            element['max'] = element['value'] + random.randint(1, 101)

    def _create_mock_elements(self, num_elements: int):
        elements = []
        for x in range(num_elements):
            elements.append({
                'stream': 'test',
                'element': 'elem%d' % x,
                'value': '--',
                'min': '--',
                'max': '--',
                'id': x
            })

        return elements


def create_app():
    r = Visualizer()
    return r.create_dev_app()


def main():  # pragma: no cover
    r = Visualizer()
    r.start()


if __name__ == "__main__":
    main()



import click
import asyncio
import argparse

import joule.errors
from joule import errors
from joule.cli.config import Config, pass_config
from joule.utilities import timestamp_to_human, human_to_timestamp
from joule.client.builtins.mean_filter import MeanFilter


@click.command(name="mean")
@click.argument("source")
@click.argument("destination")
@click.option("-w", "--window", help="window size", required=True, type=int)
@click.option("--start", help="timestamp or descriptive string")
@click.option("--end", help="timestamp or descriptive string")
@pass_config
def mean(config: Config, start, end, source, destination, window):
    """Apply a moving average (mean) filter."""
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run(start, end, source, destination, window, config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(start, end, source, destination, window, node):
    # --- RETRIEVE SOURCE STREAMS ---
    # prints out error messages if the source streams do not exist
    try:
        source_stream: joule.DataStream = await node.data_stream_get(source)
    except joule.errors.ApiError as e:
        raise click.ClickException(f"Source stream {source}: {str(e)}")

    # --- CHECK FOR DESTINATION STREAM ---
    destination_width = len(source_stream.elements)
    dest_exists = True
    dest_stream = None  # this is set below, or created later
    try:
        dest_stream = await node.data_stream_get(destination)
        if len(dest_stream.elements) != destination_width:
            raise click.ClickException(f"Destination must have {destination_width} elements")
        if not dest_stream.datatype.startswith('float'):
            raise click.ClickException(f"Destination must be a float datatype")

    except joule.errors.ApiError as e:
        dest_exists = False
    click.echo("Mean filter with window size %d" % window)
    click.echo(f"Source Stream: \n\t{source}")
    if dest_exists:
        click.echo(f"Destination Stream: \n\t{destination}")
    else:
        click.echo(f"Creating Destination: \n\t{destination}")

    if not click.confirm("Proceed?"):
        click.echo("Cancelled")
        return

    if not dest_exists:
        # create the destination
        dest_path = destination.split('/')[:-1]
        dest_name = destination.split('/')[-1]

        dest_stream = joule.api.DataStream(name=dest_name,
                                           elements=source_stream.elements,
                                           datatype='float32')
        try:
            dest_stream = await node.data_stream_create(dest_stream, '/'.join(dest_path))
        except joule.errors.ApiError as e:
            raise click.ClickException(f"Creating destination: {str(e)}")

    # --- READY TO GO ---
    for interval in await node.data_intervals(source_stream, start=start, end=end):
        start = interval[0]
        end = interval[1]
        click.echo("Processing [%s] -> [%s]" % (timestamp_to_human(start), timestamp_to_human(end)))
        mean_filter = MeanFilter()
        inputs = {'input': await node.data_read(source_stream, start, end)}
        outputs = {'output': await node.data_write(dest_stream, start, end)}
        args = argparse.Namespace(window=window)
        await mean_filter.run(args, inputs, outputs)
        await outputs['output'].close()



from joule.models.annotation import Annotation
from joule.models.element import Element
from joule.models.event_stream import EventStream
from joule.models.data_stream import DataStream
from joule.models.folder import Folder
from joule.models.module import Module
from joule.models.worker import Worker
from joule.models.proxy import Proxy
from joule.models.master import Master
from joule.models.follower import Follower
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo
from joule.models.data_store.errors import InsufficientDecimationError, DataError
from joule.models.data_store.nilmdb import NilmdbStore
from joule.models.data_store.event_store import EventStore
from joule.models.data_store.timescale import TimescaleStore
from joule.models.pipes.pipe import Pipe
from joule.models.meta import Base



import click
from tabulate import tabulate
from joule.cli.config import pass_config
from joule import api
from joule import errors


@click.command(name="list")
@pass_config
def node_list(config):
    """Display authorized nodes."""
    try:
        nodes = api.get_nodes()
        default_node = api.get_node()
    except errors.ApiError as e:
        raise click.ClickException(str(e))

    result = []
    default_indicator = click.style("\u25CF", fg="green")
    for node in nodes:
        if default_node.name == node.name:
            name = default_indicator + " " + node.name
        else:
            name = node.name
        result.append([name, node.url])

    click.echo("List of authorized nodes ("+default_indicator+"=default)")
    click.echo(tabulate(result,
                        headers=["Node", "URL"],
                        tablefmt="fancy_grid"))



from aiohttp import web

from joule.models.supervisor import Supervisor


async def index(request):
    supervisor: Supervisor = request.app["supervisor"]
    resp = []

    for proxy in supervisor.proxies:
        proxy_info = {
            "id": proxy.uuid,
            "name": proxy.name,
            "url": str(proxy.url)
        }
        resp.append(proxy_info)
    return web.json_response(data=resp)


async def info(request):
    supervisor: Supervisor = request.app["supervisor"]
    if 'name' in request.query:
        proxy = [p for p in supervisor.proxies if p.name == request.query['name']]
    elif 'id' in request.query:
        proxy = [p for p in supervisor.proxies if p.uuid == int(request.query['id'])]
    else:
        return web.Response(text="specify a name or id", status=400)
    if len(proxy) == 0:
        return web.Response(text="proxy does not exist", status=404)
    proxy = proxy[0]

    data = {
        "id": proxy.uuid,
        "name": proxy.name,
        "url": str(proxy.url)
    }
    return web.json_response(data)



import os
import configparser
import asyncio
import logging
import time
import argparse
import uvloop
import signal
import secrets
from aiohttp import web
import faulthandler
from sqlalchemy import create_engine, text
from sqlalchemy.orm import Session
from sqlalchemy.exc import SQLAlchemyError
from typing import List
import psutil
import dsnparse
import sys

from joule.models import (Base, Worker, config, EventStore,
                          DataStore, DataStream, pipes)
from joule.models.supervisor import Supervisor
from joule.errors import ConfigurationError, SubscriptionError
from joule.models import NilmdbStore, TimescaleStore, Follower
from joule.models.data_store.errors import DataError
from joule.api import TcpNode
from joule.services import (load_modules, load_streams, load_config)
import joule.middleware
import joule.controllers
from joule.utilities import ConnectionInfo
import ssl

log = logging.getLogger('joule')
async_log = logging.getLogger('asyncio')
async_log.setLevel(logging.WARNING)
Loop = asyncio.AbstractEventLoop
faulthandler.enable()

SQL_DIR = os.path.join(os.path.dirname(__file__), 'sql')


class Daemon(object):

    def __init__(self, my_config: config.JouleConfig):
        self.config: config.JouleConfig = my_config
        self.db: Session = None
        self.engine = None
        self.supervisor: Supervisor = None
        self.data_store: DataStore = None
        self.event_store: EventStore = None
        self.module_connection_info = None
        self.tasks: List[asyncio.Task] = []
        self.stop_requested = False

    def initialize(self):

        # make sure this is the only copy of jouled running
        pid_file = os.path.join(self.config.socket_directory, 'pid')
        if os.path.exists(pid_file):
            with open(pid_file, 'r') as f:
                pid = int(f.readline())
                if psutil.pid_exists(pid):
                    log.error("jouled is already running with pid %d" % pid)
                    sys.exit(1)
        # clear out the socket directory
        for file_name in os.listdir(self.config.socket_directory):
            path = os.path.join(self.config.socket_directory, file_name)
            os.unlink(path)
        # write our pid
        with open(pid_file, 'w') as f:
            f.write('%d\n' % os.getpid())
        os.chmod(pid_file, 0o600)

        if self.config.nilmdb_url is not None:
            self.data_store: DataStore = \
                NilmdbStore(self.config.nilmdb_url,
                            self.config.insert_period,
                            self.config.cleanup_period)
        else:
            self.data_store: DataStore = \
                TimescaleStore(self.config.database, self.config.insert_period,
                               self.config.cleanup_period)

        self.event_store: EventStore = EventStore(self.config.database)

        engine = create_engine(self.config.database, echo=False)
        self.engine = engine  # keep for erasing database if needed

        # create a database user for modules
        password = secrets.token_hex(8)
        parsed_dsn = dsnparse.parse(self.config.database)
        self.module_connection_info = ConnectionInfo(username="joule_module",
                                                     password=password,
                                                     port=parsed_dsn.port,
                                                     host=parsed_dsn.host,
                                                     database=parsed_dsn.database)
        # NOTE: joule user must be able to create a role and grant access to db
        # ALTER ROLE joule WITH CREATEROLE;
        # GRANT ALL PRIVILEGES ON DATABASE joule TO joule WITH GRANT OPTION;
        with engine.connect() as conn:
            conn.execute('CREATE SCHEMA IF NOT EXISTS data')
            conn.execute('CREATE SCHEMA IF NOT EXISTS metadata')

            # create a module user
            cmd = """
            DO $$
            BEGIN
                IF NOT EXISTS (SELECT 1 FROM pg_roles WHERE rolname = 'joule_module') THEN
                CREATE USER joule_module;
            END IF;
            END
            $$;"""
            conn.execute(cmd)
            conn.execute("ALTER ROLE joule_module WITH PASSWORD '%s'" % password)
            conn.execute("GRANT CONNECT ON DATABASE %s TO joule_module" % parsed_dsn.database)
            conn.execute("GRANT USAGE ON SCHEMA data TO joule_module")
            conn.execute("GRANT USAGE ON SCHEMA metadata TO joule_module")
            conn.execute("GRANT SELECT ON ALL TABLES IN SCHEMA metadata TO joule_module;")
            conn.execute("GRANT SELECT ON ALL TABLES IN SCHEMA data TO joule_module;")
            # test out pg_read_all_settings command
            conn.execute("show data_directory")
            # load custom functions
            for file in os.listdir(SQL_DIR):
                file_path = os.path.join(SQL_DIR, file)
                with open(file_path, 'r') as f:
                    conn.execute(text(f.read()))
            conn.execute("GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA public TO joule_module;")
        Base.metadata.create_all(engine)
        self.db = Session(bind=engine)

        # reset flags on streams, these will be set by load_modules and load_streams
        for stream in self.db.query(DataStream).all():
            stream.is_configured = False
            stream.is_destination = False
            stream.is_source = False

        # load modules, streams, and proxies, from configs
        load_streams.run(self.config.stream_directory, self.db)
        modules = load_modules.run(self.config.module_directory, self.db)

        # configure security parameters
        self.ssl_context = None
        self.cafile = ""
        if self.config.security is not None:
            self.ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
            self.ssl_context.load_cert_chain(certfile=self.config.security.certfile,
                                             keyfile=self.config.security.keyfile)
            if self.config.security.cafile != "":
                # publish the cafile location to the environment for modules to use
                os.environ["JOULE_CA_FILE"] = self.config.security.cafile
                self.cafile = self.config.security.cafile
        # configure workers
        workers = [Worker(m) for m in modules]

        def get_node(name: str):
            follower = self.db.query(Follower) \
                .filter_by(name=name) \
                .one_or_none()
            if follower is None:
                return None

            return TcpNode(follower.name, follower.location,
                           follower.key,
                           self.cafile)

        self.supervisor = Supervisor(workers, self.config.proxies, get_node)

        # save the metadata
        self.db.commit()

    async def run(self):
        # initialize streams in the data store and event store
        try:
            pool = await self.data_store.initialize(self.db.query(DataStream).all())
            await self.event_store.initialize(pool)
        except DataError as e:
            log.error("Database error: %s" % e)
            exit(1)
        # start the supervisor (runs the workers)
        await self.supervisor.start()

        # start inserters by subscribing to the streams
        inserter_tasks = []
        for stream in self.db.query(DataStream).filter(DataStream.keep_us != DataStream.KEEP_NONE):
            if not stream.is_destination:
                continue  # only subscribe to active streams
            try:
                task = asyncio.create_task(self._spawn_inserter(stream))
                task.set_name("Daemon: spawn_inserter [%s]" % stream.name)
                inserter_tasks.append(task)
            except SubscriptionError as e:
                logging.warning(e)
        inserter_task_grp = asyncio.gather(*inserter_tasks)
        # start the API server
        middlewares = [
            joule.middleware.sql_rollback,
            joule.middleware.authorize(
                exemptions=joule.controllers.insecure_routes),
        ]
        app = web.Application(middlewares=middlewares)

        app['module-connection-info'] = self.module_connection_info
        app['supervisor'] = self.supervisor
        app['data-store'] = self.data_store
        app['event-store'] = self.event_store
        app['db'] = self.db
        # used to tell master's this node's info

        # if the API is proxied the base_uri
        # will be retrieved from the X-Api-Base-Uri header
        app['base_uri'] = ""

        # if the API is proxied the port
        # will be retrieved from the X-Api-Port header
        app['name'] = self.config.name
        app['port'] = self.config.port

        # note, if the API is proxied the scheme
        # will be retrieved from the X-Api-Scheme header
        if self.ssl_context is None:
            app['scheme'] = 'http'
        else:
            app['scheme'] = 'https'

        # for acting as a client when accessing remote streams and joining other nodes
        app['cafile'] = self.cafile

        app.add_routes(joule.controllers.routes)
        runner = web.AppRunner(app)

        await runner.setup()
        if self.config.ip_address is not None:
            site = web.TCPSite(runner, self.config.ip_address,
                               self.config.port,
                               ssl_context=self.ssl_context)
            await site.start()

        sock_file = os.path.join(self.config.socket_directory, 'api')
        sock_site = web.UnixSite(runner, sock_file)
        await sock_site.start()
        os.chmod(sock_file, 0o600)

        # sleep and check for stop condition
        while not self.stop_requested:
            await asyncio.sleep(0.5)

        # clean everything up
        await self.supervisor.stop()
        inserter_task_grp.cancel()
        try:
            await inserter_task_grp
        except asyncio.CancelledError:
            pass
        await self.data_store.close()
        try:
            await asyncio.wait_for(runner.shutdown(), 5)
            await asyncio.wait_for(runner.cleanup(), 5)
        except asyncio.TimeoutError:
            log.warning("unclean server shutdown, subscribed clients?")
        self.db.close()

    def stop(self):
        self.stop_requested = True

    async def _spawn_inserter(self, stream: DataStream):
        while True:
            pipe = pipes.LocalPipe(layout=stream.layout, name='inserter:%s' % stream.name)
            unsubscribe = self.supervisor.subscribe(stream, pipe)
            task = None
            try:
                task = await self.data_store.spawn_inserter(stream, pipe)
                await task
                break  # inserter terminated, program is closing
            except DataError as e:
                msg = "stream [%s]: %s" % (stream.name, str(e))
                await self.supervisor.restart_producer(stream, msg=msg)
            except asyncio.CancelledError:
                if task is not None:
                    task.cancel()
                break
            unsubscribe()


def main(argv=None):
    parser = argparse.ArgumentParser("Joule Daemon")
    parser.add_argument("--config", default="/etc/joule/main.conf")
    args = parser.parse_args(argv)
    log.addFilter(LogDedupFilter())
    logging.basicConfig(
        format='%(asctime)s %(levelname)s:%(message)s',
        level=logging.WARNING)
    if args.config is not None:
        if os.path.isfile(args.config) is False:
            log.error("Invalid configuration: cannot load file [%s]" % args.config)
            exit(1)
    parser = configparser.ConfigParser()
    parser.read(args.config)
    my_config = None
    try:
        my_config = load_config.run(custom_values=parser)
    except ConfigurationError as e:
        log.error("Invalid configuration: %s" % e)
        exit(1)

    # uvloop uses libuv which does not support
    # connections to abstract namespace sockets
    # https://github.com/joyent/libuv/issues/1486

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    daemon = Daemon(my_config)
    try:
        daemon.initialize()
    except SQLAlchemyError as e:
        print("""
        Error initializing database, ensure user 'joule' has sufficient permissions:
        From a shell run
        $> sudo -u postgres psql
        postgres=# ALTER ROLE joule WITH CREATEROLE REPLICATION;
        postgres=# GRANT ALL PRIVILEGES ON DATABASE joule TO joule WITH GRANT OPTION;
        postgres=# GRANT pg_read_all_settings TO joule;
        """)
        loop.close()
        exit(1)

    loop.add_signal_handler(signal.SIGINT, daemon.stop)
    loop.add_signal_handler(signal.SIGTERM, daemon.stop)

    async def debugger():
        task_list = []
        while True:
            for t in asyncio.all_tasks():
                name = t.get_name()

                if "Task" in name:
                    if "aiohttp" in str(t.get_stack()):
                        t.set_name("aiohttp")
                    elif "StreamReader.read" in str(t.get_coro()):
                        t.set_name("StreamReader.read")
                    else:
                        t.set_name(f"UNK {t.get_coro()}")

                if t not in task_list:
                    task_list.append(t)
                    # print(f"New Task[{id(t)}]: {t.get_name()}")

            for t in task_list:
                if t.done():
                    # print(f"DONE: {t.get_name()}: ", end="")
                    # if t.cancelled():
                    #    print("cancelled")
                    if not t.cancelled():
                        exception = t.exception()
                        if exception is not None:
                            print("Got exception", t.exception())
                            #print("----Cancelling Daemon----")
                            #daemon_task.cancel()
                        # else:
                        #    print("completed")
                    task_list.remove(t)
                # else:
                #    print(f"RUNNING: {t.get_name()}")

            await asyncio.sleep(2)

    debug = loop.create_task(debugger())
    daemon_task = loop.create_task(daemon.run())
    daemon_task.set_name("daemon")
    debug.set_name("debugger")
    loop.run_until_complete(daemon_task)
    debug.cancel()
    try:
        loop.run_until_complete(debug)
    except asyncio.CancelledError:
        pass
    loop.close()

    # clear out the socket directory
    for file_name in os.listdir(my_config.socket_directory):
        path = os.path.join(my_config.socket_directory, file_name)
        os.unlink(path)
    exit(0)


class LogDedupFilter:
    def __init__(self, name='', max_gap=5):
        self.max_gap = max_gap
        self.first_repeat = True
        self.last_time = 0
        self.last_msg = None

    def filter(self, record):
        if self.last_msg is None:
            self.last_msg = record.msg
            return True
        if self.last_msg == record.msg:
            # same log entry
            now = time.time()
            prev = self.last_time
            self.last_msg = record.msg
            self.last_time = now
            if now - prev < self.max_gap:
                if self.first_repeat:
                    record.msg = "[...repeats]"
                    self.first_repeat = False
                    return True
                else:
                    return False  # suppress
            return True  # far enough apart

        self.last_time = 0
        self.last_msg = record.msg
        self.first_repeat = True
        return True


if __name__ == "__main__":
    main()



class ConfigurationError(Exception):
    """
    Error setting up an object due to incorrect configuration
    """
    pass


class SubscriptionError(Exception):
    """
    Error subscribing to a stream that is not available
    """
    pass


class ConnectionError(Exception):
    """
    Error connecting to a Joule server
    """
    pass


class DataError(Exception):
    """
    Error generated by backend storage (nilmdb or timescale)
    """
    pass


class DecimationError(DataError):
    """
    Requested data is not sufficiently decimated
    """
    pass


class ApiError(Exception):
    """
    Error generated by an API call. Catches all API related errors.
    """
    pass


class StreamNotFound(ApiError):
    """
    Type of APIError. Requested stream does not exist
    """
    pass


class EmptyPipeError(ApiError):
    """
    Type of APIError. Attempt to read from an empty pipe
    """
    pass



from joule.models.supervisor import Supervisor
from aiohttp import web
import logging

log = logging.getLogger("joule")


async def auth(request: web.Request):
    supervisor: Supervisor = request.app["supervisor"]

    try:
        app_id = request.headers["X-App-Id"]
        if len(app_id) < 2:
            raise ValueError()
        app_type = app_id[0]
        uuid = int(app_id[1:])
    except ValueError:
        return web.Response(text="invalid id", status=400)

    # rebuild the URL for the module
    if app_type == 'm':  # module
        socket = supervisor.get_module_socket(uuid)
        if socket is None:
            return web.Response(text="id not found", status=404)
        proxy_path = "http://unix:%s:/" % socket

    else:  # proxy
        target_url = supervisor.get_proxy_url(uuid)
        if target_url is None:
            return web.Response(text="id not found", status=404)
        proxy_path = str(target_url)

    return web.Response(headers={'X-Proxy-Path': proxy_path})


async def index(request: web.Request):
    supervisor: Supervisor = request.app["supervisor"]
    apps = []
    for worker in supervisor.workers:
        if not worker.is_app:
            continue
        apps.append({'id': 'm%d' % worker.uuid,
                     'name': worker.name})
    for proxy in supervisor.proxies:
        apps.append({'id': 'p%d' % proxy.uuid,
                     'name': proxy.name})
    return web.json_response(apps)



import configparser
import os
from typing import Dict
import logging

Configurations = Dict[str, configparser.ConfigParser]
logger = logging.getLogger('joule')


def load_configs(path: str) -> Configurations:
    configs: Configurations = {}
    for file in os.listdir(path):
        if not file.endswith(".conf"):
            continue
        file_path = os.path.join(path, file)
        config = configparser.ConfigParser()
        try:
            # can only happen if joule does not have read permissions
            if len(config.read(file_path)) != 1:
                logger.error("Cannot read file [%s]" % file_path)
                continue
            configs[file] = config
        except configparser.Error as e:
            logger.error("Configuration file error: %r" % e)
    return configs



from joule.errors import EmptyPipeError


class PipeError(Exception):
    """base class for numpypipe exceptions"""


# inherit from both for backwards compatibility
class EmptyPipe(PipeError, EmptyPipeError):
    pass



from aiohttp import web
import datetime

from sqlalchemy.orm import Session
from sqlalchemy.exc import SQLAlchemyError

from joule.models.annotation import Annotation, from_json
from joule.models.data_stream import DataStream
from joule.models import folder
from joule.errors import ApiError


async def index(request):
    db: Session = request.app["db"]
    # specify stream_ids as array
    # optionally specify start and end
    # parse time bounds
    start = None
    end = None
    try:
        if 'start' in request.query:
            ts = int(request.query['start'])
            start = datetime.datetime.utcfromtimestamp(ts / 1e6)
        if 'end' in request.query:
            ts = int(request.query['end'])
            end = datetime.datetime.utcfromtimestamp(ts / 1e6)
    except ValueError:
        return web.Response(text="[start] and [end] must be microsecond utc timestamps", status=400)
    if (('stream_id' not in request.query) and
            ('stream_path' not in request.query)):
        return web.Response(text="must specify at least one stream_id or stream_path", status=400)

    response = []
    stream_ids = []
    if "stream_id" in request.query:
        stream_ids = request.query.getall("stream_id")
    if "stream_path" in request.query:
        for path in request.query.getall("stream_path"):
            stream = folder.find_stream_by_path(path, db, stream_type=DataStream)
            if stream is None:
                return web.Response(text="stream [%s] does not exist" % path, status=404)
            stream_ids.append(stream.id)

    for stream_id in stream_ids:
        annotations = db.query(Annotation).filter_by(stream_id=stream_id)
        if start is not None:
            annotations = annotations.filter(Annotation.start >= start)
        if end is not None:
            annotations = annotations.filter(Annotation.start <= end)
        data = [a.to_json() for a in annotations]
        response += [a.to_json() for a in annotations]

    return web.json_response(response)


async def update(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    if 'id' in body:
        my_annotation = db.query(Annotation).get(body["id"])
    else:
        return web.Response(text="specify an id", status=400)
    if my_annotation is None:
        return web.Response(text="annotation does not exist", status=404)
    my_annotation.update_attributes(body)
    if my_annotation.title is None or my_annotation.title == '':
        return web.Response(text="annotation title is reqiured")
    db.commit()
    return web.json_response(my_annotation.to_json())


async def create(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    if 'stream_id' in body:
        my_stream = db.query(DataStream).get(body["stream_id"])
    elif 'stream_path' in body:
        my_stream = folder.find_stream_by_path(body['stream_path'], db, stream_type=DataStream)
    else:
        return web.Response(text="specify a stream_id", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)

    try:
        my_annotation = from_json(body)
        if my_annotation.title is None or my_annotation.title == '':
            raise ApiError("title is reqiured")
    except ApiError as e:
        return web.Response(text=str(e), status=400)

    my_annotation.stream = my_stream
    db.add(my_annotation)
    db.commit()
    return web.json_response(my_annotation.to_json())


async def delete(request):
    db: Session = request.app["db"]
    if 'id' in request.query:
        my_annotation = db.query(Annotation).get(request.query["id"])
    else:
        return web.Response(text="specify an id", status=400)
    if my_annotation is None:
        return web.Response(text="annotation does not exist", status=404)
    db.delete(my_annotation)
    db.commit()
    return web.Response(text="ok")


async def delete_all(request):
    db: Session = request.app["db"]

    if "stream_id" in request.query:
        stream_id = request.query["stream_id"]
        my_stream = db.query(DataStream).get(stream_id)
    elif "stream_path" in request.query:
        path = request.query["stream_path"]
        my_stream = folder.find_stream_by_path(path, db, stream_type=DataStream)
    else:
        return web.Response(text="must specify either stream_id or stream_path", status=400)

    start = None
    end = None
    try:
        if 'start' in request.query:
            ts = int(request.query['start'])
            start = datetime.datetime.utcfromtimestamp(ts / 1e6)
        if 'end' in request.query:
            ts = int(request.query['end'])
            end = datetime.datetime.utcfromtimestamp(ts / 1e6)
    except ValueError:
        return web.Response(text="[start] and [end] must be microsecond utc timestamps", status=400)

    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)

    annotations = db.query(Annotation).filter_by(stream_id=my_stream.id)

    if start is not None:
        annotations = annotations.filter(Annotation.start >= start)
    if end is not None:
        annotations = annotations.filter(Annotation.start <= end)

    for annotation in annotations:
        db.delete(annotation)
    db.commit()
    return web.Response(text="ok")



#from joule.services.parse_pipe_config import parse_inline_config

#from joule.api.data_stream import DataStream, Element


def build_stream(name, inline_config: str):# -> DataStream:
    assert(False, 'deprecated')
    s = DataStream()
    s.name = name
    (datatype, names) = parse_inline_config(inline_config)
    s.datatype = datatype.name.lower(),
    i = 0
    for name in names:
        e = Element()
        e.name = name
        e.index = i
        s.elements.append(e)
    return s



from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, Enum, Float, String, Boolean, ForeignKey
import configparser
import enum
from typing import Optional, Dict, TYPE_CHECKING

from joule.models.meta import Base
from joule.errors import ConfigurationError

if TYPE_CHECKING:  # pragma: no cover
    from joule.models.folder import DataStream


class Element(Base):
    """
    Attributes:
        name (str): element name
        index (int): order of element in the stream
        units (str): data unit (eg Watts, Volts, etc)
        plottable (bool): whether the data can be visualized as a time series
        offset (float): linear scaling y=mx+b, only applied to Lumen visualizations
        scale_factor (float): linear scaling only applied to Lumen visualizations
        default_min (float): fix lower limit on autoscaling in Lumen visualizations
        default_max (float): fix upper limit on autoscaling in Lumen visualizations
        display_type (Element.DISPLAYTYPE): visualization type
    """
    __tablename__ = 'element'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    index: int = Column(Integer, nullable=False)
    name: str = Column(String)
    units: str = Column(String)
    plottable: bool = Column(Boolean)

    class DISPLAYTYPE(enum.Enum):
        DISCRETE = enum.auto()
        CONTINUOUS = enum.auto()
        EVENT = enum.auto()

    display_type: DISPLAYTYPE = Column(Enum(DISPLAYTYPE),
                                       default=DISPLAYTYPE.CONTINUOUS)

    offset: float = Column(Float, default=0, nullable=False)
    scale_factor: float = Column(Float, default=1.0, nullable=False)
    default_max: Optional[float] = Column(Float, default=None)
    default_min: Optional[float] = Column(Float, default=None)
    stream_id: int = Column(Integer, ForeignKey('metadata.stream.id'))
    stream: 'DataStream' = relationship("DataStream", back_populates="elements")

    def __repr__(self):
        return "<Element(name='%s', units='%s', display_type=%s)>" % (self.name, self.units, self.display_type)

    def to_json(self):
        """

        Returns: Dictionary of Element attributes

        """
        if self.display_type is None:
            # make up a default value
            self.display_type = Element.DISPLAYTYPE.CONTINUOUS
        return {
            'id': self.id,
            'index': self.index,
            'name': self.name,
            'units': self.units,
            'plottable': self.plottable,
            'display_type': self.display_type.name,
            'offset': self.offset,
            'scale_factor': self.scale_factor,
            'default_max': self.default_max,
            'default_min': self.default_min
        }

    def to_nilmdb_metadata(self):
        return {
            'column': self.index,
            'name': self.name,
            'units': self.units,
            'scale_factor': self.scale_factor,
            'offset': self.offset,
            'plottable': True,  # TODO: check why elements have NULL fields here?
            'discrete': False,
            'default_min': self.default_min,
            'default_max': self.default_max
        }

    def update_attributes(self, attrs: Dict):
        if 'name' in attrs:
            self.name = validate_name(attrs['name'])
        if 'units' in attrs:
            self.units = attrs['units']
        try:
            if 'plottable' in attrs:
                self.plottable = bool(attrs['plottable'])
            if 'offset' in attrs:
                self.offset = float(attrs['offset'])
            if 'scale_factor' in attrs:
                self.scale_factor = float(attrs['scale_factor'])
            if 'display_type' in attrs:
                self.display_type = validate_type(attrs['display_type'])
            if 'default_max' in attrs:
                x = attrs['default_max']
                if x is not None:
                    x = float(x)
                self.default_max = x
            if 'default_min' in attrs:
                x = attrs['default_min']
                if x is not None:
                    x = float(x)
                self.default_min = x
        except ValueError as e:
            raise ConfigurationError("Invalid element configuration: %r" % e)
        if self.default_max is not None and self.default_min is not None:
            if self.default_max <= self.default_min:
                raise ConfigurationError("[default_min] must be > [default_max]")


def from_json(data: Dict) -> Element:
    return Element(id=data["id"],
                   index=data["index"],
                   name=validate_name(data["name"]),
                   units=data["units"],
                   plottable=data["plottable"],
                   display_type=validate_type(data["display_type"].upper()),
                   offset=data["offset"],
                   scale_factor=data["scale_factor"],
                   default_max=data["default_max"],
                   default_min=data["default_min"])


def from_config(config: configparser.ConfigParser) -> Element:
    name = validate_name(config["name"])
    display_type = validate_type(config.get("display_type", fallback="continuous"))
    units = config.get("units", fallback=None)
    plottable = _get_bool("plottable", config, True)
    offset = _get_float("offset", config, 0.0)
    scale_factor = _get_float("scale_factor", config, 1.0)
    default_max = _get_float("default_max", config, None)
    default_min = _get_float("default_min", config, None)
    # make sure min<=max if set
    if ((default_max is not None and default_min is not None) and
            (default_min >= default_max)):
        raise ConfigurationError("set default_min<default_max or omit for autoscale")
    return Element(name=name, units=units, plottable=plottable,
                   display_type=display_type, offset=offset,
                   scale_factor=scale_factor,
                   default_max=default_max,
                   default_min=default_min)


def from_nilmdb_metadata(config: Dict, default_index) -> Element:
    # cannot assume any structure to the config dict
    default_config = {
        "name": "elem%d" % default_index,
        "units": None,
        "plottable": True,
        "offset": 0,
        "scale_factor": 1.0,
        "default_max": None,
        "default_min": None,
        "index": default_index,
        "discrete": False
    }
    merged_config = {**default_config, **config}
    display_type = Element.DISPLAYTYPE.CONTINUOUS
    if merged_config['discrete']:
        display_type = Element.DISPLAYTYPE.EVENT

    return Element(name=merged_config["name"],
                   display_type=display_type,
                   units=merged_config['units'],
                   plottable=merged_config['plottable'],
                   offset=merged_config['offset'],
                   scale_factor=merged_config['scale_factor'],
                   default_max=merged_config['default_max'],
                   default_min=merged_config['default_min'],
                   index=merged_config['column'])


def validate_name(name: str) -> str:
    if name == "":
        raise ConfigurationError("missing name")
    return name


def validate_type(display_type: str) -> Element.DISPLAYTYPE:
    try:
        return Element.DISPLAYTYPE[display_type.upper()]
    except KeyError as e:
        valid_types = ", ".join([m.name.lower() for m in Element.DISPLAYTYPE])
        raise ConfigurationError("invalid display_type [%s], choose from [%s]" %
                                 (display_type, valid_types)) from e


def _get_bool(setting: str, config: configparser.ConfigParser, default: bool):
    try:
        return config.getboolean(setting, default)
    except ValueError as e:
        raise ConfigurationError("[%s] invalid value, use True/False" % setting) from e


def _get_float(setting: str, config: configparser.ConfigParser, default):
    try:
        if config[setting] == "None":
            return default
        return config.getfloat(setting, default)
    except KeyError:
        return default
    except ValueError as e:
        raise ConfigurationError("[%s] invalid value, must be a number" %
                                 setting) from e
